# Copyright (c) Ralph Meijer.
# See LICENSE for details

"""
Tests for L{wokkel.xmppim}.
"""

from __future__ import division, absolute_import

from twisted.internet import defer
from twisted.trial import unittest
from twisted.words.protocols.jabber import error
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.xmlstream import toResponse
from twisted.words.xish import domish, utility

from wokkel import xmppim
from wokkel.generic import ErrorStanza, parseXml
from wokkel.test.helpers import TestableRequestHandlerMixin, XmlStreamStub

NS_XML = "http://www.w3.org/XML/1998/namespace"
NS_ROSTER = "jabber:iq:roster"


class PresenceClientProtocolTest(unittest.TestCase):
    def setUp(self):
        self.output = []
        self.protocol = xmppim.PresenceClientProtocol()
        self.protocol.parent = self

    def send(self, obj):
        self.output.append(obj)

    def test_unavailableDirected(self):
        """
        Test sending of directed unavailable presence broadcast.
        """

        self.protocol.unavailable(JID("user@example.com"))
        presence = self.output[-1]
        self.assertEquals("presence", presence.name)
        self.assertEquals(None, presence.uri)
        self.assertEquals("user@example.com", presence.getAttribute("to"))
        self.assertEquals("unavailable", presence.getAttribute("type"))

    def test_unavailableWithStatus(self):
        """
        Test sending of directed unavailable presence broadcast with status.
        """

        self.protocol.unavailable(JID("user@example.com"), {None: "Disconnected"})
        presence = self.output[-1]
        self.assertEquals("presence", presence.name)
        self.assertEquals(None, presence.uri)
        self.assertEquals("user@example.com", presence.getAttribute("to"))
        self.assertEquals("unavailable", presence.getAttribute("type"))
        self.assertEquals("Disconnected", str(presence.status))

    def test_unavailableBroadcast(self):
        """
        Test sending of unavailable presence broadcast.
        """

        self.protocol.unavailable(None)
        presence = self.output[-1]
        self.assertEquals("presence", presence.name)
        self.assertEquals(None, presence.uri)
        self.assertEquals(None, presence.getAttribute("to"))
        self.assertEquals("unavailable", presence.getAttribute("type"))

    def test_unavailableBroadcastNoEntityParameter(self):
        """
        Test sending of unavailable presence broadcast by not passing entity.
        """

        self.protocol.unavailable()
        presence = self.output[-1]
        self.assertEquals("presence", presence.name)
        self.assertEquals(None, presence.uri)
        self.assertEquals(None, presence.getAttribute("to"))
        self.assertEquals("unavailable", presence.getAttribute("type"))


class AvailabilityPresenceTest(unittest.TestCase):

    def test_fromElement(self):
        xml = """<presence from='user@example.org' to='user@example.com'>
                   <show>chat</show>
                   <status>Let's chat!</status>
                   <priority>50</priority>
                 </presence>
              """

        presence = xmppim.AvailabilityPresence.fromElement(parseXml(xml))
        self.assertEquals(JID("user@example.org"), presence.sender)
        self.assertEquals(JID("user@example.com"), presence.recipient)
        self.assertTrue(presence.available)
        self.assertEquals("chat", presence.show)
        self.assertEquals({None: "Let's chat!"}, presence.statuses)
        self.assertEquals(50, presence.priority)


class PresenceProtocolTest(unittest.TestCase):
    """
    Tests for L{xmppim.PresenceProtocol}
    """

    def setUp(self):
        self.output = []
        self.protocol = xmppim.PresenceProtocol()
        self.protocol.parent = self
        self.protocol.xmlstream = utility.EventDispatcher()
        self.protocol.connectionInitialized()

    def send(self, obj):
        self.output.append(obj)

    def test_errorReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type="error"/>"""

        def errorReceived(error):
            xmppim.PresenceProtocol.errorReceived(self.protocol, error)
            try:
                self.assertIsInstance(error, ErrorStanza)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.errorReceived = errorReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_availableReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence/>"""

        def availableReceived(presence):
            xmppim.PresenceProtocol.availableReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.AvailabilityPresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.availableReceived = availableReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_unavailableReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type='unavailable'/>"""

        def unavailableReceived(presence):
            xmppim.PresenceProtocol.unavailableReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.AvailabilityPresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.unavailableReceived = unavailableReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_subscribeReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type='subscribe'/>"""

        def subscribeReceived(presence):
            xmppim.PresenceProtocol.subscribeReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.SubscriptionPresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.subscribeReceived = subscribeReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_unsubscribeReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type='unsubscribe'/>"""

        def unsubscribeReceived(presence):
            xmppim.PresenceProtocol.unsubscribeReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.SubscriptionPresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.unsubscribeReceived = unsubscribeReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_subscribedReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type='subscribed'/>"""

        def subscribedReceived(presence):
            xmppim.PresenceProtocol.subscribedReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.SubscriptionPresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.subscribedReceived = subscribedReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_unsubscribedReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type='unsubscribed'/>"""

        def unsubscribedReceived(presence):
            xmppim.PresenceProtocol.unsubscribedReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.SubscriptionPresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.unsubscribedReceived = unsubscribedReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_probeReceived(self):
        """
        Incoming presence stanzas are parsed and dispatched.
        """
        xml = """<presence type='probe'/>"""

        def probeReceived(presence):
            xmppim.PresenceProtocol.probeReceived(self.protocol, presence)
            try:
                self.assertIsInstance(presence, xmppim.ProbePresence)
            except:
                d.errback()
            else:
                d.callback(None)

        d = defer.Deferred()
        self.protocol.probeReceived = probeReceived
        self.protocol.xmlstream.dispatch(parseXml(xml))
        return d

    def test_available(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.available(
            JID("user@example.com"), show="chat", status="Talk to me!", priority=50
        )
        element = self.output[-1]
        self.assertEquals("user@example.com", element.getAttribute("to"))
        self.assertIdentical(None, element.getAttribute("type"))
        self.assertEquals("chat", str(element.show))
        self.assertEquals("Talk to me!", str(element.status))
        self.assertEquals("50", str(element.priority))

    def test_availableLanguages(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.available(
            JID("user@example.com"),
            show="chat",
            statuses={None: "Talk to me!", "nl": "Praat met me!"},
            priority=50,
        )
        element = self.output[-1]
        self.assertEquals("user@example.com", element.getAttribute("to"))
        self.assertIdentical(None, element.getAttribute("type"))
        self.assertEquals("chat", str(element.show))

        statuses = {}
        for status in element.elements():
            if status.name == "status":
                lang = status.getAttribute((NS_XML, "lang"))
                statuses[lang] = str(status)

        self.assertIn(None, statuses)
        self.assertEquals("Talk to me!", statuses[None])
        self.assertIn("nl", statuses)
        self.assertEquals("Praat met me!", statuses["nl"])
        self.assertEquals("50", str(element.priority))

    def test_availableSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.available(JID("user@example.com"), sender=JID("user@example.org"))
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_unavailableDirected(self):
        """
        Test sending of directed unavailable presence broadcast.
        """

        self.protocol.unavailable(JID("user@example.com"))
        element = self.output[-1]
        self.assertEquals("presence", element.name)
        self.assertEquals(None, element.uri)
        self.assertEquals("user@example.com", element.getAttribute("to"))
        self.assertEquals("unavailable", element.getAttribute("type"))

    def test_unavailableWithStatus(self):
        """
        Test sending of directed unavailable presence broadcast with status.
        """

        self.protocol.unavailable(JID("user@example.com"), {None: "Disconnected"})
        element = self.output[-1]
        self.assertEquals("presence", element.name)
        self.assertEquals(None, element.uri)
        self.assertEquals("user@example.com", element.getAttribute("to"))
        self.assertEquals("unavailable", element.getAttribute("type"))
        self.assertEquals("Disconnected", str(element.status))

    def test_unavailableBroadcast(self):
        """
        Test sending of unavailable presence broadcast.
        """

        self.protocol.unavailable(None)
        element = self.output[-1]
        self.assertEquals("presence", element.name)
        self.assertEquals(None, element.uri)
        self.assertEquals(None, element.getAttribute("to"))
        self.assertEquals("unavailable", element.getAttribute("type"))

    def test_unavailableBroadcastNoRecipientParameter(self):
        """
        Test sending of unavailable presence broadcast by not passing entity.
        """

        self.protocol.unavailable()
        element = self.output[-1]
        self.assertEquals("presence", element.name)
        self.assertEquals(None, element.uri)
        self.assertEquals(None, element.getAttribute("to"))
        self.assertEquals("unavailable", element.getAttribute("type"))

    def test_unavailableSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.unavailable(JID("user@example.com"), sender=JID("user@example.org"))
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_subscribeSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.subscribe(JID("user@example.com"), sender=JID("user@example.org"))
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_unsubscribeSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.unsubscribe(JID("user@example.com"), sender=JID("user@example.org"))
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_subscribedSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.subscribed(JID("user@example.com"), sender=JID("user@example.org"))
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_unsubscribedSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.unsubscribed(
            JID("user@example.com"), sender=JID("user@example.org")
        )
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_probeSender(self):
        """
        It should be possible to pass a sender address.
        """
        self.protocol.probe(JID("user@example.com"), sender=JID("user@example.org"))
        element = self.output[-1]
        self.assertEquals("user@example.org", element.getAttribute("from"))


class RosterItemTest(unittest.TestCase):
    """
    Tests for L{xmppim.RosterItem}.
    """

    def test_toElement(self):
        """
        A roster item has the correct namespace/name, lacks unset attributes.
        """
        item = xmppim.RosterItem(JID("user@example.org"))
        element = item.toElement()
        self.assertEqual("item", element.name)
        self.assertEqual(NS_ROSTER, element.uri)
        self.assertFalse(element.hasAttribute("subscription"))
        self.assertFalse(element.hasAttribute("ask"))
        self.assertEqual("", element.getAttribute("name", ""))
        self.assertFalse(element.hasAttribute("approved"))
        self.assertEquals(0, len(list(element.elements())))

    def test_toElementMinimal(self):
        """
        A bare roster item only has a jid attribute.
        """
        item = xmppim.RosterItem(JID("user@example.org"))
        element = item.toElement()
        self.assertEqual("user@example.org", element.getAttribute("jid"))

    def test_toElementSubscriptionNone(self):
        """
        A roster item with no subscription has no subscription attribute.
        """
        item = xmppim.RosterItem(
            JID("user@example.org"), subscriptionTo=False, subscriptionFrom=False
        )
        element = item.toElement()
        self.assertIdentical(None, element.getAttribute("subscription"))

    def test_toElementSubscriptionTo(self):
        """
        A roster item with subscriptionTo set has subscription 'to'.
        """
        item = xmppim.RosterItem(
            JID("user@example.org"), subscriptionTo=True, subscriptionFrom=False
        )
        element = item.toElement()
        self.assertEqual("to", element.getAttribute("subscription"))

    def test_toElementSubscriptionFrom(self):
        """
        A roster item with subscriptionFrom set has subscription 'to'.
        """
        item = xmppim.RosterItem(
            JID("user@example.org"), subscriptionTo=False, subscriptionFrom=True
        )
        element = item.toElement()
        self.assertEqual("from", element.getAttribute("subscription"))

    def test_toElementSubscriptionBoth(self):
        """
        A roster item with mutual subscription has subscription 'both'.
        """
        item = xmppim.RosterItem(
            JID("user@example.org"), subscriptionTo=True, subscriptionFrom=True
        )
        element = item.toElement()
        self.assertEqual("both", element.getAttribute("subscription"))

    def test_toElementSubscriptionRemove(self):
        """
        A roster item with remove set has subscription 'remove'.
        """
        item = xmppim.RosterItem(JID("user@example.org"))
        item.remove = True
        element = item.toElement()
        self.assertEqual("remove", element.getAttribute("subscription"))

    def test_toElementAsk(self):
        """
        A roster item with pendingOut set has subscription 'ask'.
        """
        item = xmppim.RosterItem(JID("user@example.org"))
        item.pendingOut = True
        element = item.toElement()
        self.assertEqual("subscribe", element.getAttribute("ask"))

    def test_toElementName(self):
        """
        A roster item's name is rendered to the 'name' attribute.
        """
        item = xmppim.RosterItem(JID("user@example.org"), name="Joe User")
        element = item.toElement()
        self.assertEqual("Joe User", element.getAttribute("name"))

    def test_toElementGroups(self):
        """
        A roster item's groups are rendered as 'group' child elements.
        """
        groups = set(["Friends", "Jabber"])
        item = xmppim.RosterItem(JID("user@example.org"), groups=groups)

        element = item.toElement()
        foundGroups = set()
        for child in element.elements():
            if child.uri == NS_ROSTER and child.name == "group":
                foundGroups.add(str(child))

        self.assertEqual(groups, foundGroups)

    def test_toElementApproved(self):
        """
        A pre-approved subscription for a roster item has an 'approved' flag.
        """
        item = xmppim.RosterItem(JID("user@example.org"))
        item.approved = True
        element = item.toElement()
        self.assertEqual("true", element.getAttribute("approved"))

    def test_fromElementMinimal(self):
        """
        A minimal roster item has a reference to the JID of the contact.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertEqual(JID("test@example.org"), item.entity)
        self.assertEqual("", item.name)
        self.assertFalse(item.subscriptionTo)
        self.assertFalse(item.subscriptionFrom)
        self.assertFalse(item.pendingOut)
        self.assertFalse(item.approved)
        self.assertEqual(set(), item.groups)

    def test_fromElementName(self):
        """
        A roster item may have an optional name.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  name="Test User"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertEqual("Test User", item.name)

    def test_fromElementGroups(self):
        """
        A roster item may have one or more groups.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org">
              <group>Friends</group>
              <group>Twisted</group>
            </item>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertIn("Twisted", item.groups)
        self.assertIn("Friends", item.groups)

    def test_fromElementSubscriptionNone(self):
        """
        Subscription 'none' sets both attributes to False.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  subscription="none"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertFalse(item.remove)
        self.assertFalse(item.subscriptionTo)
        self.assertFalse(item.subscriptionFrom)

    def test_fromElementSubscriptionTo(self):
        """
        Subscription 'to' sets the corresponding attribute to True.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  subscription="to"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertFalse(item.remove)
        self.assertTrue(item.subscriptionTo)
        self.assertFalse(item.subscriptionFrom)

    def test_fromElementSubscriptionFrom(self):
        """
        Subscription 'from' sets the corresponding attribute to True.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  subscription="from"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertFalse(item.remove)
        self.assertFalse(item.subscriptionTo)
        self.assertTrue(item.subscriptionFrom)

    def test_fromElementSubscriptionBoth(self):
        """
        Subscription 'both' sets both attributes to True.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  subscription="both"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertFalse(item.remove)
        self.assertTrue(item.subscriptionTo)
        self.assertTrue(item.subscriptionFrom)

    def test_fromElementSubscriptionRemove(self):
        """
        Subscription 'remove' sets the remove attribute.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  subscription="remove"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertTrue(item.remove)

    def test_fromElementPendingOut(self):
        """
        The ask attribute, if set to 'subscription', means pending out.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  ask="subscribe"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertTrue(item.pendingOut)

    def test_fromElementApprovedTrue(self):
        """
        The approved attribute (true) signals a pre-approved subscription.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  approved="true"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertTrue(item.approved)

    def test_fromElementApproved1(self):
        """
        The approved attribute (1) signals a pre-approved subscription.
        """

        xml = """
            <item xmlns="jabber:iq:roster"
                  jid="test@example.org"
                  approved="1"/>
        """

        item = xmppim.RosterItem.fromElement(parseXml(xml))
        self.assertTrue(item.approved)


class RosterRequestTest(unittest.TestCase):
    """
    Tests for L{xmppim.RosterRequest}.
    """

    def test_fromElement(self):
        """
        A bare roster request is parsed and missing information is None.
        """
        xml = """
            <iq type='get' to='this@example.org/Home' from='this@example.org'>
              <query xmlns='jabber:iq:roster'/>
            </iq>
        """

        request = xmppim.RosterRequest.fromElement(parseXml(xml))
        self.assertEqual("get", request.stanzaType)
        self.assertEqual(JID("this@example.org/Home"), request.recipient)
        self.assertEqual(JID("this@example.org"), request.sender)
        self.assertEqual(None, request.item)
        self.assertEqual(None, request.version)

    def test_fromElementItem(self):
        """
        If an item is present, parse it and put it in the request item.
        """
        xml = """
            <iq type='set' to='this@example.org/Home' from='this@example.org'>
              <query xmlns='jabber:iq:roster'>
                <item jid='user@example.org'/>
              </query>
            </iq>
        """

        request = xmppim.RosterRequest.fromElement(parseXml(xml))
        self.assertNotIdentical(None, request.item)
        self.assertEqual(JID("user@example.org"), request.item.entity)

    def test_fromElementVersion(self):
        """
        If a ver attribute is present, put it in the request version.
        """
        xml = """
            <iq type='set' to='this@example.org/Home' from='this@example.org'>
              <query xmlns='jabber:iq:roster' ver='ver72'>
                <item jid='user@example.org'/>
              </query>
            </iq>
        """
        request = xmppim.RosterRequest.fromElement(parseXml(xml))
        self.assertEqual("ver72", request.version)

    def test_fromElementVersionEmpty(self):
        """
        The ver attribute may be empty.
        """
        xml = """
            <iq type='get' to='this@example.org/Home' from='this@example.org'>
              <query xmlns='jabber:iq:roster' ver=''/>
            </iq>
        """
        request = xmppim.RosterRequest.fromElement(parseXml(xml))
        self.assertEqual("", request.version)

    def test_toElement(self):
        """
        A roster request has a query element in the roster namespace.
        """
        request = xmppim.RosterRequest()
        element = request.toElement()
        children = element.elements()
        child = next(children)
        self.assertEqual(NS_ROSTER, child.uri)
        self.assertEqual("query", child.name)

    def test_toElementItem(self):
        """
        If an item is set, it is rendered as a child of the query.
        """
        request = xmppim.RosterRequest()
        request.item = xmppim.RosterItem(JID("user@example.org"))
        element = request.toElement()
        children = element.query.elements()
        child = next(children)
        self.assertEqual(NS_ROSTER, child.uri)
        self.assertEqual("item", child.name)


class FakeClient(object):
    """
    Fake client stream manager for roster tests.
    """

    def __init__(self, xmlstream, jid):
        self.xmlstream = xmlstream
        self.jid = jid

    def request(self, request):
        element = request.toElement()
        self.xmlstream.send(element)
        return defer.Deferred()

    def addHandler(self, handler):
        handler.makeConnection(self.xmlstream)
        handler.connectionInitialized()

    def test_toElementVersion(self):
        """
        If the roster version is set, a 'ver' attribute is added.
        """
        request = xmppim.RosterRequest()
        request.version = "ver72"
        element = request.toElement()
        self.assertEqual("ver72", element.query.getAttribute("ver"))

    def test_toElementVersionEmpty(self):
        """
        If the roster version is the empty string, it should add 'ver', too.
        """
        request = xmppim.RosterRequest()
        request.version = ""
        element = request.toElement()
        self.assertEqual("", element.query.getAttribute("ver"))


class RosterClientProtocolTest(unittest.TestCase, TestableRequestHandlerMixin):
    """
    Tests for L{xmppim.RosterClientProtocol}.
    """

    def setUp(self):
        self.stub = XmlStreamStub()
        self.client = FakeClient(self.stub.xmlstream, JID("this@example.org"))
        self.service = xmppim.RosterClientProtocol()
        self.service.setHandlerParent(self.client)

    def test_setItem(self):
        """
        Setting a roster item renders the item and sends it out.
        """
        item = xmppim.RosterItem(
            JID("test@example.org"), name="Joe User", groups=set(["Friends", "Jabber"])
        )
        d = self.service.setItem(item)

        # Inspect outgoing iq request

        iq = self.stub.output[-1]
        self.assertEqual("set", iq.getAttribute("type"))
        self.assertNotIdentical(None, iq.query)
        self.assertEqual(NS_ROSTER, iq.query.uri)

        children = list(
            domish.generateElementsQNamed(iq.query.children, "item", NS_ROSTER)
        )
        self.assertEqual(1, len(children))
        child = children[0]
        self.assertEqual("test@example.org", child["jid"])
        self.assertIdentical(None, child.getAttribute("subscription"))

        # Fake successful response

        response = toResponse(iq, "result")
        d.callback(response)
        return d

    def test_setItemIgnoreAttributes(self):
        """
        Certain attributes should be rendered for roster set.
        """
        item = xmppim.RosterItem(
            JID("test@example.org"),
            subscriptionTo=True,
            subscriptionFrom=False,
            name="Joe User",
            groups=set(["Friends", "Jabber"]),
        )
        item.pendingOut = True
        item.approved = True
        d = self.service.setItem(item)

        # Inspect outgoing iq request

        iq = self.stub.output[-1]
        self.assertEqual("set", iq.getAttribute("type"))
        self.assertNotIdentical(None, iq.query)
        self.assertEqual(NS_ROSTER, iq.query.uri)

        children = list(
            domish.generateElementsQNamed(iq.query.children, "item", NS_ROSTER)
        )
        self.assertEqual(1, len(children))
        child = children[0]
        self.assertIdentical(None, child.getAttribute("ask"))
        self.assertIdentical(None, child.getAttribute("approved"))
        self.assertIdentical(None, child.getAttribute("subscription"))

        # Fake successful response

        response = toResponse(iq, "result")
        d.callback(response)
        return d

    def test_removeItem(self):
        """
        Removing a roster item is setting an item with subscription C{remove}.
        """
        d = self.service.removeItem(JID("test@example.org"))

        # Inspect outgoing iq request

        iq = self.stub.output[-1]
        self.assertEqual("set", iq.getAttribute("type"))
        self.assertNotIdentical(None, iq.query)
        self.assertEqual(NS_ROSTER, iq.query.uri)

        children = list(
            domish.generateElementsQNamed(iq.query.children, "item", NS_ROSTER)
        )
        self.assertEqual(1, len(children))
        child = children[0]
        self.assertEqual("test@example.org", child["jid"])
        self.assertEqual("remove", child.getAttribute("subscription"))

        # Fake successful response

        response = toResponse(iq, "result")
        d.callback(response)
        return d

    def test_getRoster(self):
        """
        A request for the roster is sent out and the response is parsed.
        """

        def cb(roster):
            self.assertIn(JID("user@example.org"), roster)
            self.assertIdentical(None, getattr(roster, "version"))

        d = self.service.getRoster()
        d.addCallback(cb)

        # Inspect outgoing iq request

        iq = self.stub.output[-1]
        self.assertEqual("get", iq.getAttribute("type"))
        self.assertNotIdentical(None, iq.query)
        self.assertEqual(NS_ROSTER, iq.query.uri)
        self.assertFalse(iq.query.hasAttribute("ver"))

        # Fake successful response
        response = toResponse(iq, "result")
        query = response.addElement((NS_ROSTER, "query"))
        item = query.addElement("item")
        item["jid"] = "user@example.org"

        d.callback(response)
        return d

    def test_getRosterVer(self):
        """
        A request for the roster with version passes the version on.
        """

        def cb(roster):
            self.assertEqual("ver96", getattr(roster, "version"))

        d = self.service.getRoster(version="ver72")
        d.addCallback(cb)

        # Inspect outgoing iq request

        iq = self.stub.output[-1]
        self.assertEqual("ver72", iq.query.getAttribute("ver"))

        # Fake successful response
        response = toResponse(iq, "result")
        query = response.addElement((NS_ROSTER, "query"))
        query["ver"] = "ver96"
        item = query.addElement("item")
        item["jid"] = "user@example.org"

        d.callback(response)
        return d

    def test_getRosterVerEmptyResult(self):
        """
        An empty response is returned as None.
        """

        def cb(response):
            self.assertIdentical(None, response)

        d = self.service.getRoster(version="ver72")
        d.addCallback(cb)

        # Inspect outgoing iq request

        iq = self.stub.output[-1]

        # Fake successful response
        response = toResponse(iq, "result")
        d.callback(response)
        return d

    def test_setReceived(self):
        """
        A roster set push causes setReceived.
        """
        xml = """
          <iq type='set'>
            <query xmlns='jabber:iq:roster'>
              <item jid='user@example.org'/>
            </query>
          </iq>
        """

        requests = []

        def setReceived(request):
            requests.append(request)

        def cb(result):
            self.assertEqual(1, len(requests), "setReceived was not called")
            self.assertEqual(JID("user@example.org"), requests[0].item.entity)

        self.service.setReceived = setReceived

        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d

    def test_setReceivedOtherSource(self):
        """
        Roster pushes can be sent from other entities, too, ignore them.
        """
        xml = """
          <iq type='set' to='this@example.org/Home' from='other@example.org'>
            <query xmlns='jabber:iq:roster'>
              <item jid='user@example.org'/>
            </query>
          </iq>
        """

        def cb(result):
            self.assertEquals("service-unavailable", result.condition)

        d = self.handleRequest(xml)
        self.assertFailure(d, error.StanzaError)
        d.addCallback(cb)
        return d

    def test_setReceivedOtherSourceAllowed(self):
        """
        Roster pushes can be sent from other entities, allow them.
        """
        xml = """
          <iq type='set' to='this@example.org/Home' from='other@example.org'>
            <query xmlns='jabber:iq:roster'>
              <item jid='user@example.org'/>
            </query>
          </iq>
        """

        self.service.allowAnySender = True
        requests = []

        def setReceived(request):
            requests.append(request)

        def cb(result):
            self.assertEqual(1, len(requests), "setReceived was not called")

        self.service.setReceived = setReceived

        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d

    def test_setReceivedOtherSourceIgnored(self):
        """
        Roster pushes can be sent from other entities, allow them.
        """
        xml = """
          <iq type='set' to='this@example.org/Home' from='bad@example.org'>
            <query xmlns='jabber:iq:roster'>
              <item jid='user@example.org'/>
            </query>
          </iq>
        """

        self.service.allowAnySender = True

        def setReceived(request):
            if request.sender == JID("bad@example.org"):
                raise xmppim.RosterPushIgnored()

        def cb(result):
            self.assertEquals("service-unavailable", result.condition)

        self.service.setReceived = setReceived

        d = self.handleRequest(xml)
        self.assertFailure(d, error.StanzaError)
        d.addCallback(cb)
        return d

    def test_removeReceived(self):
        """
        A roster remove push causes removeReceived.
        """
        xml = """
          <iq type='set'>
            <query xmlns='jabber:iq:roster'>
              <item jid='user@example.org' subscription='remove'/>
            </query>
          </iq>
        """

        requests = []

        def removeReceived(request):
            requests.append(request)

        def cb(result):
            self.assertEqual(1, len(requests), "removeReceived was not called")
            self.assertEqual(JID("user@example.org"), requests[0].item.entity)

        self.service.removeReceived = removeReceived

        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
XMPP Component Service.

This provides an XMPP server that accepts External Components connections
and accepts and initiates server-to-server connections for the specified
domain(s).
"""

from __future__ import division, absolute_import

from twisted.application import service, strports
from twisted.python import usage
from twisted.words.protocols.jabber import component
from wokkel import server


class Options(usage.Options):

    optParameters = [
        (
            "component-port",
            None,
            "tcp:5347:interface=127.0.0.1",
            "Port components connect to",
        ),
        ("component-secret", None, "secret", "Secret components use to connect"),
        ("server-port", None, "tcp:5269", "Port other servers connect to"),
        ("server-secret", None, None, "Shared secret for dialback verification"),
    ]

    optFlags = [
        ("verbose", "v", "Log traffic"),
    ]

    def __init__(self):
        usage.Options.__init__(self)
        self["domains"] = set()

    def opt_domain(self, domain):
        """
        Domain to accept server connections for. Repeat for more domains.
        """
        self["domains"].add(domain)

    def postOptions(self):
        if not self["domains"]:
            raise usage.UsageError("Need at least one domain")


def makeService(config):
    s = service.MultiService()

    router = component.Router()

    # Set up the XMPP server service

    serverService = server.ServerService(router, secret=config["server-secret"])
    serverService.domains = config["domains"]
    serverService.logTraffic = config["verbose"]

    # Hook up XMPP server-to-server service
    s2sFactory = server.XMPPS2SServerFactory(serverService)
    s2sFactory.logTraffic = config["verbose"]
    s2sService = strports.service(config["server-port"], s2sFactory)
    s2sService.setServiceParent(s)

    # Hook up XMPP external server-side component service
    cFactory = component.XMPPComponentServerFactory(router, config["component-secret"])

    cFactory.logTraffic = config["verbose"]
    cServer = strports.service(config["component-port"], cFactory)
    cServer.setServiceParent(s)

    return s



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for {wokkel.data_form}.
"""

from __future__ import division, absolute_import

from zope.interface import verify
from zope.interface.common.mapping import IIterableMapping

from twisted.trial import unittest
from twisted.words.xish import domish
from twisted.words.protocols.jabber import jid

from wokkel import data_form

NS_X_DATA = "jabber:x:data"


class OptionTest(unittest.TestCase):
    """
    Tests for L{data_form.Option}.
    """

    def test_toElement(self):
        """
        An option is an option element with a value child with the option value.
        """
        option = data_form.Option("value")
        element = option.toElement()

        self.assertEqual("option", element.name)
        self.assertEqual(NS_X_DATA, element.uri)
        self.assertEqual(NS_X_DATA, element.value.uri)
        self.assertEqual("value", str(element.value))
        self.assertFalse(element.hasAttribute("label"))

    def test_toElementLabel(self):
        """
        A label is rendered as an attribute on the option element.
        """
        option = data_form.Option("value", "label")
        element = option.toElement()

        self.assertEqual("option", element.name)
        self.assertEqual(NS_X_DATA, element.uri)
        self.assertEqual(NS_X_DATA, element.value.uri)
        self.assertEqual("value", str(element.value))
        self.assertEqual("label", element["label"])

    def test_fromElement(self):
        """
        An option has a child element with the option value.
        """
        element = domish.Element((NS_X_DATA, "option"))
        element.addElement("value", content="value")
        option = data_form.Option.fromElement(element)

        self.assertEqual("value", option.value)
        self.assertIdentical(None, option.label)

    def test_fromElementLabel(self):
        """
        An option label is an attribute on the option element.
        """

        element = domish.Element((NS_X_DATA, "option"))
        element.addElement("value", content="value")
        element["label"] = "label"
        option = data_form.Option.fromElement(element)

        self.assertEqual("label", option.label)

    def test_fromElementNoValue(self):
        """
        An option MUST have a value.
        """
        element = domish.Element((NS_X_DATA, "option"))
        self.assertRaises(data_form.Error, data_form.Option.fromElement, element)

    def test_repr(self):
        """
        The representation of an Option is equal to how it is created.
        """
        option = data_form.Option("value", "label")
        self.assertEqual("""Option('value', 'label')""", repr(option))


class FieldTest(unittest.TestCase):
    """
    Tests for L{data_form.Field}.
    """

    def test_basic(self):
        """
        Test basic field initialization.
        """
        field = data_form.Field(var="test")
        self.assertEqual("text-single", field.fieldType)
        self.assertEqual("test", field.var)

    def test_labelAndOptions(self):
        """
        The label should be set, even if there are options with labels as dict.
        """
        field = data_form.Field(
            label="test", options={"test2": "test 2", "test3": "test 3"}
        )
        self.assertEqual("test", field.label)

    def test_repr(self):
        """
        The repr of a field should be equal to its initialization.
        """
        field = data_form.Field(
            "list-single",
            var="test",
            label="label",
            desc="desc",
            required=True,
            value="test",
            options=[data_form.Option("test")],
        )
        self.assertEqual(
            """Field(fieldType='list-single', """
            """var='test', label='label', """
            """desc='desc', required=True, """
            """values=['test'], """
            """options=[Option('test')])""",
            repr(field),
        )

    def test_toElement(self):
        """
        Test rendering to a DOM.
        """
        field = data_form.Field(var="test")
        element = field.toElement()

        self.assertTrue(domish.IElement.providedBy(element))
        self.assertEquals("field", element.name)
        self.assertEquals(NS_X_DATA, element.uri)
        self.assertEquals("text-single", element.getAttribute("type", "text-single"))
        self.assertEquals("test", element["var"])
        self.assertEquals([], element.children)

    def test_toElementTypeNotTextSingle(self):
        """
        Always render the field type, if different from text-single.
        """
        field = data_form.Field("hidden", var="test")
        element = field.toElement()

        self.assertEquals("hidden", element.getAttribute("type"))

    def test_toElementSingleValue(self):
        """
        A single value should yield only one value element.
        """
        field = data_form.Field("list-multi", var="test", value="test")
        element = field.toElement()

        children = list(element.elements())
        self.assertEqual(1, len(children))

    def test_toElementMultipleValues(self):
        """
        A field with no type and multiple values should render all values.
        """
        field = data_form.Field("list-multi", var="test", values=["test", "test2"])
        element = field.toElement()

        children = list(element.elements())
        self.assertEqual(2, len(children))

    def test_toElementAsForm(self):
        """
        Always render the field type, if asForm is True.
        """
        field = data_form.Field(var="test")
        element = field.toElement(True)

        self.assertEquals("text-single", element.getAttribute("type"))

    def test_toElementOptions(self):
        """
        Test rendering to a DOM with options.
        """
        field = data_form.Field("list-single", var="test")
        field.options = [data_form.Option("option1"), data_form.Option("option2")]
        element = field.toElement(True)

        self.assertEqual(2, len(element.children))

    def test_toElementLabel(self):
        """
        Test rendering to a DOM with a label.
        """
        field = data_form.Field(var="test", label="my label")
        element = field.toElement(True)

        self.assertEqual("my label", element.getAttribute("label"))

    def test_toElementDescription(self):
        """
        Test rendering to a DOM with options.
        """
        field = data_form.Field(var="test", desc="My desc")
        element = field.toElement(True)

        self.assertEqual(1, len(element.children))
        child = element.children[0]
        self.assertEqual("desc", child.name)
        self.assertEqual(NS_X_DATA, child.uri)
        self.assertEqual("My desc", str(child))

    def test_toElementRequired(self):
        """
        Test rendering to a DOM with options.
        """
        field = data_form.Field(var="test", required=True)
        element = field.toElement(True)

        self.assertEqual(1, len(element.children))
        child = element.children[0]
        self.assertEqual("required", child.name)
        self.assertEqual(NS_X_DATA, child.uri)

    def test_toElementJID(self):
        """
        A JID value should render to text.
        """
        field = data_form.Field(
            fieldType="jid-single", var="test", value=jid.JID("test@example.org")
        )
        element = field.toElement()
        self.assertEqual("test@example.org", str(element.value))

    def test_toElementJIDTextSingle(self):
        """
        A JID value should render to text if field type is text-single.
        """
        field = data_form.Field(
            fieldType="text-single", var="test", value=jid.JID("test@example.org")
        )
        element = field.toElement()
        self.assertEqual("test@example.org", str(element.value))

    def test_toElementBoolean(self):
        """
        A boolean value should render to text.
        """
        field = data_form.Field(fieldType="boolean", var="test", value=True)
        element = field.toElement()
        self.assertEqual("true", str(element.value))

    def test_toElementBooleanTextSingle(self):
        """
        A boolean value should render to text if the field type is text-single.
        """
        field = data_form.Field(var="test", value=True)
        element = field.toElement()
        self.assertEqual("true", str(element.value))

    def test_toElementNoType(self):
        """
        A field with no type should not have a type attribute.
        """
        field = data_form.Field(None, var="test", value="test")
        element = field.toElement()
        self.assertFalse(element.hasAttribute("type"))

    def test_toElementNoTypeMultipleValues(self):
        """
        A field with no type and multiple values should render all values.
        """
        field = data_form.Field(None, var="test", values=["test", "test2"])
        element = field.toElement()

        self.assertFalse(element.hasAttribute("type"))
        children = list(element.elements())
        self.assertEqual(2, len(children))

    def test_typeCheckNoFieldName(self):
        """
        A field not of type fixed must have a var.
        """
        field = data_form.Field(fieldType="list-single")
        self.assertRaises(data_form.FieldNameRequiredError, field.typeCheck)

    def test_typeCheckTooManyValues(self):
        """
        Expect an exception if too many values are set, depending on type.
        """
        field = data_form.Field(
            fieldType="list-single", var="test", values=["value1", "value2"]
        )
        self.assertRaises(data_form.TooManyValuesError, field.typeCheck)

    def test_typeCheckBooleanFalse(self):
        """
        Test possible False values for a boolean field.
        """
        field = data_form.Field(fieldType="boolean", var="test")

        for value in (False, 0, "0", "false", "False", []):
            field.value = value
            field.typeCheck()
            self.assertIsInstance(field.value, bool)
            self.assertFalse(field.value)

    def test_typeCheckBooleanTrue(self):
        """
        Test possible True values for a boolean field.
        """
        field = data_form.Field(fieldType="boolean", var="test")

        for value in (True, 1, "1", "true", "True", ["test"]):
            field.value = value
            field.typeCheck()
            self.assertIsInstance(field.value, bool)
            self.assertTrue(field.value)

    def test_typeCheckBooleanBad(self):
        """
        A bad value for a boolean field should raise a ValueError
        """
        field = data_form.Field(fieldType="boolean", var="test")
        field.value = "test"
        self.assertRaises(ValueError, field.typeCheck)

    def test_typeCheckJID(self):
        """
        The value of jid field should be a JID or coercable to one.
        """
        field = data_form.Field(
            fieldType="jid-single", var="test", value=jid.JID("test@example.org")
        )
        field.typeCheck()

    def test_typeCheckJIDString(self):
        """
        The string value of jid field should be coercable into a JID.
        """
        field = data_form.Field(
            fieldType="jid-single", var="test", value="test@example.org"
        )
        field.typeCheck()
        self.assertEquals(jid.JID("test@example.org"), field.value)

    def test_typeCheckJIDBad(self):
        """
        An invalid JID string should raise an exception.
        """
        field = data_form.Field(
            fieldType="jid-single", var="test", value="test@@example.org"
        )
        self.assertRaises(jid.InvalidFormat, field.typeCheck)

    def test_fromElementType(self):
        element = domish.Element((NS_X_DATA, "field"))
        element["type"] = "fixed"
        field = data_form.Field.fromElement(element)
        self.assertEquals("fixed", field.fieldType)

    def test_fromElementNoType(self):
        element = domish.Element((NS_X_DATA, "field"))
        field = data_form.Field.fromElement(element)
        self.assertEquals(None, field.fieldType)

    def test_fromElementValueTextSingle(self):
        """
        Parsed text-single field values should be of type L{str}.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element["type"] = "text-single"
        element.addElement("value", content="text")
        field = data_form.Field.fromElement(element)
        self.assertEquals("text", field.value)

    def test_fromElementValueJID(self):
        """
        Parsed jid-single field values should be of type L{str}.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element["type"] = "jid-single"
        element.addElement("value", content="user@example.org")
        field = data_form.Field.fromElement(element)
        self.assertEquals("user@example.org", field.value)

    def test_fromElementValueJIDMalformed(self):
        """
        Parsed jid-single field values should be of type L{str}.

        No validation should be done at this point, so invalid JIDs should
        also be passed as-is.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element["type"] = "jid-single"
        element.addElement("value", content="@@")
        field = data_form.Field.fromElement(element)
        self.assertEquals("@@", field.value)

    def test_fromElementValueBoolean(self):
        """
        Parsed boolean field values should be of type L{str}.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element["type"] = "boolean"
        element.addElement("value", content="false")
        field = data_form.Field.fromElement(element)
        self.assertEquals("false", field.value)

    def test_fromElementDesc(self):
        """
        Field descriptions are in a desc child element.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element.addElement("desc", content="My description")
        field = data_form.Field.fromElement(element)
        self.assertEqual("My description", field.desc)

    def test_fromElementOption(self):
        """
        Field descriptions are in a desc child element.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element.addElement("option").addElement("value", content="option1")
        element.addElement("option").addElement("value", content="option2")
        field = data_form.Field.fromElement(element)
        self.assertEqual(2, len(field.options))

    def test_fromElementRequired(self):
        """
        Required fields have a required child element.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element.addElement("required")
        field = data_form.Field.fromElement(element)
        self.assertTrue(field.required)

    def test_fromElementChildOtherNamespace(self):
        """
        Child elements from another namespace are ignored.
        """
        element = domish.Element((NS_X_DATA, "field"))
        element["var"] = "test"
        element.addElement(("myns", "value"))
        field = data_form.Field.fromElement(element)

        self.assertIdentical(None, field.value)

    def test_fromDict(self):
        """
        A named field with a value can be created by providing a dictionary.
        """
        fieldDict = {"var": "test", "value": "text"}
        field = data_form.Field.fromDict(fieldDict)
        self.assertEqual("test", field.var)
        self.assertEqual("text", field.value)

    def test_fromDictFieldType(self):
        """
        The field type is set using the key 'type'.
        """
        fieldDict = {"type": "boolean"}
        field = data_form.Field.fromDict(fieldDict)
        self.assertEqual("boolean", field.fieldType)

    def test_fromDictOptions(self):
        """
        The field options are set using the key 'options'.

        The options are represented as a dictionary keyed by option,
        with the optional label as value.
        """
        fieldDict = {"options": {"value1": "label1", "value2": "label2"}}
        field = data_form.Field.fromDict(fieldDict)
        self.assertEqual(2, len(field.options))
        options = {}
        for option in field.options:
            options[option.value] = option.label

        self.assertEqual(options, fieldDict["options"])


class FormTest(unittest.TestCase):
    """
    Tests for L{data_form.Form}.
    """

    def test_formType(self):
        """
        A form has a type.
        """

        form = data_form.Form("result")
        self.assertEqual("result", form.formType)

    def test_toElement(self):
        """
        The toElement method returns a form's DOM representation.
        """
        form = data_form.Form("result")
        element = form.toElement()

        self.assertTrue(domish.IElement.providedBy(element))
        self.assertEquals("x", element.name)
        self.assertEquals(NS_X_DATA, element.uri)
        self.assertEquals("result", element["type"])
        self.assertEquals([], element.children)

    def test_toElementTitle(self):
        """
        A title is rendered as a child element with the title as CDATA.
        """
        form = data_form.Form("form", title="Bot configuration")
        element = form.toElement()

        elements = list(element.elements())
        self.assertEqual(1, len(elements))
        title = elements[0]
        self.assertEqual("title", title.name)
        self.assertEqual(NS_X_DATA, title.uri)
        self.assertEqual("Bot configuration", str(title))

    def test_toElementInstructions(self):
        """
        Instructions are rendered as child elements with CDATA.
        """
        form = data_form.Form("form", instructions=["Fill out this form!"])
        element = form.toElement()

        elements = list(element.elements())
        self.assertEqual(1, len(elements))
        instructions = elements[0]
        self.assertEqual("instructions", instructions.name)
        self.assertEqual(NS_X_DATA, instructions.uri)
        self.assertEqual("Fill out this form!", str(instructions))

    def test_toElementInstructionsMultiple(self):
        """
        Instructions render as one element per instruction, in order.
        """
        form = data_form.Form("form", instructions=["Fill out this form!", "no really"])
        element = form.toElement()

        elements = list(element.elements())
        self.assertEqual(2, len(elements))
        instructions1 = elements[0]
        instructions2 = elements[1]
        self.assertEqual("instructions", instructions1.name)
        self.assertEqual(NS_X_DATA, instructions1.uri)
        self.assertEqual("Fill out this form!", str(instructions1))
        self.assertEqual("instructions", instructions2.name)
        self.assertEqual(NS_X_DATA, instructions2.uri)
        self.assertEqual("no really", str(instructions2))

    def test_toElementFormType(self):
        """
        The form type is rendered as a hidden field with name FORM_TYPE.
        """
        form = data_form.Form("form", formNamespace="jabber:bot")
        element = form.toElement()

        elements = list(element.elements())
        self.assertEqual(1, len(elements))
        formTypeField = elements[0]
        self.assertEqual("field", formTypeField.name)
        self.assertEqual(NS_X_DATA, formTypeField.uri)
        self.assertEqual("FORM_TYPE", formTypeField["var"])
        self.assertEqual("hidden", formTypeField["type"])
        self.assertEqual("jabber:bot", str(formTypeField.value))

    def test_toElementFields(self):
        """
        Fields are rendered as child elements, in order.
        """
        fields = [
            data_form.Field("fixed", value="Section 1"),
            data_form.Field("text-single", var="botname", label="The name of your bot"),
            data_form.Field(
                "text-multi", var="description", label="Helpful description of your bot"
            ),
            data_form.Field("boolean", var="public", label="Public bot?", required=True),
        ]
        form = data_form.Form("form", fields=fields)
        element = form.toElement()

        elements = list(element.elements())
        self.assertEqual(4, len(elements))
        for field in elements:
            self.assertEqual("field", field.name)
            self.assertEqual(NS_X_DATA, field.uri)

        # Check order
        self.assertEqual("fixed", elements[0]["type"])
        self.assertEqual("botname", elements[1]["var"])
        self.assertEqual("description", elements[2]["var"])
        self.assertEqual("public", elements[3]["var"])

    def test_fromElement(self):
        """
        C{fromElement} creates a L{data_form.Form} from a DOM representation.
        """
        element = domish.Element((NS_X_DATA, "x"))
        element["type"] = "result"
        form = data_form.Form.fromElement(element)

        self.assertEquals("result", form.formType)
        self.assertEquals(None, form.title)
        self.assertEquals([], form.instructions)
        self.assertEquals({}, form.fields)

    def test_fromElementInvalidElementName(self):
        """
        Bail if the passed element does not have the correct name.
        """
        element = domish.Element((NS_X_DATA, "form"))
        self.assertRaises(Exception, data_form.Form.fromElement, element)

    def test_fromElementInvalidElementURI(self):
        """
        Bail if the passed element does not have the correct namespace.
        """
        element = domish.Element(("myns", "x"))
        self.assertRaises(Exception, data_form.Form.fromElement, element)

    def test_fromElementTitle(self):
        element = domish.Element((NS_X_DATA, "x"))
        element.addElement("title", content="My title")
        form = data_form.Form.fromElement(element)

        self.assertEquals("My title", form.title)

    def test_fromElementInstructions(self):
        element = domish.Element((NS_X_DATA, "x"))
        element.addElement("instructions", content="instruction")
        form = data_form.Form.fromElement(element)

        self.assertEquals(["instruction"], form.instructions)

    def test_fromElementInstructions2(self):
        element = domish.Element((NS_X_DATA, "x"))
        element.addElement("instructions", content="instruction 1")
        element.addElement("instructions", content="instruction 2")
        form = data_form.Form.fromElement(element)

        self.assertEquals(["instruction 1", "instruction 2"], form.instructions)

    def test_fromElementOneField(self):
        element = domish.Element((NS_X_DATA, "x"))
        element.addElement("field")
        form = data_form.Form.fromElement(element)

        self.assertEquals(1, len(form.fieldList))
        self.assertNotIn("field", form.fields)

    def test_fromElementTwoFields(self):
        element = domish.Element((NS_X_DATA, "x"))
        element.addElement("field")["var"] = "field1"
        element.addElement("field")["var"] = "field2"
        form = data_form.Form.fromElement(element)

        self.assertEquals(2, len(form.fieldList))
        self.assertIn("field1", form.fields)
        self.assertEquals("field1", form.fieldList[0].var)
        self.assertIn("field2", form.fields)
        self.assertEquals("field2", form.fieldList[1].var)

    def test_fromElementFormType(self):
        """
        The form type is a hidden field named FORM_TYPE.
        """
        element = domish.Element((NS_X_DATA, "x"))
        field = element.addElement("field")
        field["var"] = "FORM_TYPE"
        field["type"] = "hidden"
        field.addElement("value", content="myns")
        form = data_form.Form.fromElement(element)

        self.assertNotIn("FORM_TYPE", form.fields)
        self.assertEqual("myns", form.formNamespace)

    def test_fromElementFormTypeNotHidden(self):
        """
        A non-hidden field named FORM_TYPE does not set the form type.
        """
        element = domish.Element((NS_X_DATA, "x"))
        field = element.addElement("field")
        field["var"] = "FORM_TYPE"
        field.addElement("value", content="myns")
        form = data_form.Form.fromElement(element)

        self.assertIn("FORM_TYPE", form.fields)
        self.assertIdentical(None, form.formNamespace)

    def test_fromElementChildOtherNamespace(self):
        """
        Child elements from another namespace are ignored.
        """
        element = domish.Element((NS_X_DATA, "x"))
        element["type"] = "result"
        field = element.addElement(("myns", "field"))
        field["var"] = "test"
        form = data_form.Form.fromElement(element)

        self.assertEqual(0, len(form.fields))

    def test_repr(self):
        """
        The repr of a form should be equal to its initialization.
        """
        form = data_form.Form(
            "form",
            title="title",
            instructions=["instr"],
            formNamespace="myns",
            fields=[data_form.Field("fixed", value="test")],
        )
        self.assertEqual(
            """Form(formType='form', title='title', """
            """instructions=['instr'], formNamespace='myns', """
            """fields=[Field(fieldType='fixed', """
            """values=['test'])])""",
            repr(form),
        )

    def test_addField(self):
        """
        A field should occur in fieldList.
        """
        form = data_form.Form("result")
        field = data_form.Field("fixed", value="Section 1")
        form.addField(field)
        self.assertEqual([field], form.fieldList)

    def test_addFieldTwice(self):
        """
        Fields occur in fieldList in the order they were added.
        """
        form = data_form.Form("result")
        field1 = data_form.Field("fixed", value="Section 1")
        field2 = data_form.Field("fixed", value="Section 2")
        form.addField(field1)
        form.addField(field2)
        self.assertEqual([field1, field2], form.fieldList)

    def test_addFieldNotNamed(self):
        """
        A non-named field should not occur in fields.
        """
        form = data_form.Form("result")
        field = data_form.Field("fixed", value="Section 1")
        form.addField(field)
        self.assertEqual({}, form.fields)

    def test_addFieldNamed(self):
        """
        A named field should occur in fields.
        """
        form = data_form.Form("result")
        field = data_form.Field(var="test")
        form.addField(field)
        self.assertEqual({"test": field}, form.fields)

    def test_addFieldTwiceNamed(self):
        """
        A second named field should occur in fields.
        """
        form = data_form.Form("result")
        field1 = data_form.Field(var="test")
        field2 = data_form.Field(var="test2")
        form.addField(field2)
        form.addField(field1)
        self.assertEqual({"test": field1, "test2": field2}, form.fields)

    def test_addFieldSameName(self):
        """
        A named field cannot occur twice.
        """
        form = data_form.Form("result")
        field1 = data_form.Field(var="test", value="value")
        field2 = data_form.Field(var="test", value="value2")
        form.addField(field1)
        self.assertRaises(data_form.Error, form.addField, field2)

    def test_removeField(self):
        """
        A removed field should not occur in fieldList.
        """
        form = data_form.Form("result")
        field = data_form.Field("fixed", value="Section 1")
        form.addField(field)
        form.removeField(field)
        self.assertNotIn(field, form.fieldList)

    def test_removeFieldNamed(self):
        """
        A removed named field should not occur in fields.
        """
        form = data_form.Form("result")
        field = data_form.Field(var="test", value="test1")
        form.addField(field)
        form.removeField(field)
        self.assertNotIn("test", form.fields)

    def test_makeField(self):
        """
        Fields can be created from a dict of values and a dict of field defs.
        """
        fieldDefs = {
            "pubsub#persist_items": {
                "type": "boolean",
                "label": "Persist items to storage",
            },
            "pubsub#deliver_payloads": {
                "type": "boolean",
                "label": "Deliver payloads with event notifications",
            },
            "pubsub#creator": {
                "type": "jid-single",
                "label": "The JID of the node creator",
            },
            "pubsub#description": {
                "type": "text-single",
                "label": "A description of the node",
            },
            "pubsub#owner": {"type": "jid-single", "label": "Owner of the node"},
        }
        values = {
            "pubsub#deliver_payloads": "0",
            "pubsub#persist_items": True,
            "pubsub#description": "a great node",
            "pubsub#owner": jid.JID("user@example.org"),
            "x-myfield": ["a", "b"],
        }

        form = data_form.Form("submit")
        form.makeFields(values, fieldDefs)

        # Check that the expected fields have been created
        self.assertIn("pubsub#deliver_payloads", form.fields)
        self.assertIn("pubsub#persist_items", form.fields)
        self.assertIn("pubsub#description", form.fields)
        self.assertIn("pubsub#owner", form.fields)

        # This field is not created because there is no value for it.
        self.assertNotIn("pubsub#creator", form.fields)

        # This field is not created because it does not appear in fieldDefs
        # and filterUnknown defaults to True
        self.assertNotIn("x-myfield", form.fields)

        # Check properties the created fields
        self.assertEqual("boolean", form.fields["pubsub#deliver_payloads"].fieldType)
        self.assertEqual("0", form.fields["pubsub#deliver_payloads"].value)
        self.assertEqual(
            "Deliver payloads with event notifications",
            form.fields["pubsub#deliver_payloads"].label,
        )
        self.assertEqual(True, form.fields["pubsub#persist_items"].value)

    def test_makeFieldNotFilterUnknown(self):
        """
        Fields can be created from a dict of values and a dict of field defs.
        """
        fieldDefs = {
            "pubsub#persist_items": {
                "type": "boolean",
                "label": "Persist items to storage",
            },
        }
        values = {"x-myfield": ["a", "b"]}

        form = data_form.Form("submit")
        form.makeFields(values, fieldDefs, filterUnknown=False)

        field = form.fields["x-myfield"]
        self.assertEqual(None, field.fieldType)
        self.assertEqual(values, form.getValues())

    def test_makeFieldsUnknownTypeJID(self):
        """
        Without type, a single JID value sets field type jid-single.
        """
        values = {"pubsub#creator": jid.JID("user@example.org")}
        form = data_form.Form("result")
        form.makeFields(values)

        field = form.fields["pubsub#creator"]
        self.assertEqual(None, field.fieldType)
        self.assertEqual(values, form.getValues())

    def test_makeFieldsUnknownTypeJIDMulti(self):
        """
        Without type, multiple JID values sets field type jid-multi.
        """
        values = {
            "pubsub#contact": [jid.JID("user@example.org"), jid.JID("other@example.org")]
        }
        form = data_form.Form("result")
        form.makeFields(values)

        field = form.fields["pubsub#contact"]
        self.assertEqual(None, field.fieldType)
        self.assertEqual(values, form.getValues())

    def test_makeFieldsUnknownTypeBoolean(self):
        """
        Without type, a boolean value sets field type boolean.
        """
        values = {"pubsub#persist_items": True}
        form = data_form.Form("result")
        form.makeFields(values)

        field = form.fields["pubsub#persist_items"]
        self.assertEqual(None, field.fieldType)
        self.assertEqual(values, form.getValues())

    def test_makeFieldsUnknownTypeListMulti(self):
        """
        Without type, multiple values sets field type list-multi.
        """
        values = {"pubsub#show-values": ["chat", "online", "away"]}
        form = data_form.Form("result")
        form.makeFields(values)

        field = form.fields["pubsub#show-values"]
        self.assertEqual(None, field.fieldType)
        self.assertEqual(values, form.getValues())

    def test_interface(self):
        """
        L{Form}s act as a read-only dictionary.
        """
        form = data_form.Form("submit")
        verify.verifyObject(IIterableMapping, form)

    def test_getitem(self):
        """
        Using Form as a mapping will yield the value of fields keyed by name.
        """
        fields = [
            data_form.Field(var="botname", value="The Jabber Bot"),
            data_form.Field("boolean", var="public", value=True),
            data_form.Field("list-multi", var="features", values=["news", "search"]),
        ]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual("The Jabber Bot", form["botname"])
        self.assertTrue(form["public"])
        self.assertEqual(["news", "search"], form["features"])

    def test_getitemOneValueTypeMulti(self):
        """
        A single value for a multi-value field type is returned in a list.
        """
        fields = [data_form.Field("list-multi", var="features", values=["news"])]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual(["news"], form["features"])

    def test_getitemMultipleValuesNoType(self):
        """
        Multiple values for a field without type are returned in a list.
        """
        fields = [data_form.Field(None, var="features", values=["news", "search"])]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual(["news", "search"], form["features"])

    def test_getitemMultipleValuesTypeSingle(self):
        """
        Multiple values for a single-value field type returns the first value.
        """
        fields = [
            data_form.Field("text-single", var="features", values=["news", "search"])
        ]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual("news", form["features"])

    def test_get(self):
        """
        Getting the value of a known field succeeds.
        """
        fields = [data_form.Field(var="botname", value="The Jabber Bot")]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual("The Jabber Bot", form.get("botname"))

    def test_getUnknownNone(self):
        """
        Getting the value of a unknown field returns None.
        """
        fields = [data_form.Field(var="botname", value="The Jabber Bot")]
        form = data_form.Form("submit", fields=fields)
        self.assertIdentical(None, form.get("features"))

    def test_getUnknownDefault(self):
        """
        Getting the value of a unknown field returns specified default.
        """
        fields = [data_form.Field(var="botname", value="The Jabber Bot")]
        form = data_form.Form("submit", fields=fields)
        self.assertTrue(form.get("public", True))

    def test_contains(self):
        """
        A form contains a known field.
        """
        fields = [data_form.Field(var="botname", value="The Jabber Bot")]
        form = data_form.Form("submit", fields=fields)
        self.assertIn("botname", form)

    def test_containsNot(self):
        """
        A form does not contains an unknown field.
        """
        fields = [data_form.Field(var="botname", value="The Jabber Bot")]
        form = data_form.Form("submit", fields=fields)
        self.assertNotIn("features", form)

    def test_keys(self):
        """
        Iterating over the keys of a form yields all field names.
        """
        fields = [
            data_form.Field(var="botname", value="The Jabber Bot"),
            data_form.Field("boolean", var="public", value=True),
            data_form.Field("list-multi", var="features", values=["news", "search"]),
        ]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual(set(["botname", "public", "features"]), set(form.keys()))

    def test_values(self):
        """
        Iterating over the values of a form yields all field values.
        """
        fields = [
            data_form.Field(var="botname", value="The Jabber Bot"),
            data_form.Field("boolean", var="public", value=True),
        ]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual(set(["The Jabber Bot", True]), set(form.values()))

    def test_items(self):
        """
        Iterating over the values of a form yields all item tuples.
        """
        fields = [
            data_form.Field(var="botname", value="The Jabber Bot"),
            data_form.Field("boolean", var="public", value=True),
        ]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual(
            set([("botname", "The Jabber Bot"), ("public", True)]), set(form.items())
        )

    def test_getValues(self):
        """
        L{Form.getValues} returns a dict of all field values.
        """
        fields = [
            data_form.Field(var="botname", value="The Jabber Bot"),
            data_form.Field("boolean", var="public", value=True),
            data_form.Field("list-multi", var="features", values=["news", "search"]),
        ]
        form = data_form.Form("submit", fields=fields)
        self.assertEqual(
            {"botname": "The Jabber Bot", "public": True, "features": ["news", "search"]},
            form.getValues(),
        )

    def test_typeCheckKnownFieldChecked(self):
        """
        Known fields are type checked.
        """
        checked = []
        fieldDefs = {
            "pubsub#description": {
                "type": "text-single",
                "label": "A description of the node",
            }
        }
        form = data_form.Form("submit")
        form.addField(data_form.Field(var="pubsub#description", value="a node"))
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck(fieldDefs)

        self.assertEqual([None], checked)

    def test_typeCheckKnownFieldNoType(self):
        """
        Known fields without a type get the type of the field definition.
        """
        checked = []
        fieldDefs = {
            "pubsub#description": {
                "type": "text-single",
                "label": "A description of the node",
            }
        }
        form = data_form.Form("submit")
        form.addField(data_form.Field(None, var="pubsub#description", value="a node"))
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck(fieldDefs)

        self.assertEqual("text-single", field.fieldType)
        self.assertEqual([None], checked)

    def test_typeCheckWrongFieldType(self):
        """
        A field should have the same type as the field definition.
        """
        checked = []
        fieldDefs = {
            "pubsub#description": {
                "type": "text-single",
                "label": "A description of the node",
            }
        }
        form = data_form.Form("submit")
        form.addField(
            data_form.Field("list-single", var="pubsub#description", value="a node")
        )
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)

        self.assertRaises(TypeError, form.typeCheck, fieldDefs)
        self.assertEqual([], checked)

    def test_typeCheckDefaultTextSingle(self):
        """
        If a field definition has no type, use text-single.
        """
        checked = []
        fieldDefs = {"pubsub#description": {"label": "A description of the node"}}
        form = data_form.Form("submit")
        form.addField(
            data_form.Field("text-single", var="pubsub#description", value="a node")
        )
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck(fieldDefs)

        self.assertEqual([None], checked)

    def test_typeCheckUnknown(self):
        """
        Unknown fields are checked, not removed if filterUnknown False.
        """
        checked = []
        fieldDefs = {}
        form = data_form.Form("submit")
        form.addField(
            data_form.Field("list-single", var="pubsub#description", value="a node")
        )
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck(fieldDefs, filterUnknown=False)

        self.assertIn("pubsub#description", form.fields)
        self.assertEqual([None], checked)

    def test_typeCheckUnknownNoType(self):
        """
        Unknown fields without type are not checked.
        """
        checked = []
        fieldDefs = {}
        form = data_form.Form("submit")
        form.addField(data_form.Field(None, var="pubsub#description", value="a node"))
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck(fieldDefs, filterUnknown=False)

        self.assertIn("pubsub#description", form.fields)
        self.assertEqual([], checked)

    def test_typeCheckUnknownRemoved(self):
        """
        Unknown fields are not checked, and removed if filterUnknown True.
        """
        checked = []
        fieldDefs = {}
        form = data_form.Form("submit")
        form.addField(
            data_form.Field("list-single", var="pubsub#description", value="a node")
        )
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck(fieldDefs, filterUnknown=True)

        self.assertNotIn("pubsub#description", form.fields)
        self.assertEqual([], checked)


class FindFormTest(unittest.TestCase):
    """
    Tests for L{data_form.findForm}.
    """

    def test_findForm(self):
        element = domish.Element((None, "test"))
        theForm = data_form.Form("submit", formNamespace="myns")
        element.addChild(theForm.toElement())
        form = data_form.findForm(element, "myns")
        self.assertEqual("myns", form.formNamespace)

    def test_noFormType(self):
        element = domish.Element((None, "test"))
        otherForm = data_form.Form("submit")
        element.addChild(otherForm.toElement())
        form = data_form.findForm(element, "myns")
        self.assertIdentical(None, form)

    def test_noFormTypeCancel(self):
        """
        Cancelled forms don't have a FORM_TYPE field, the first is returned.
        """
        element = domish.Element((None, "test"))
        cancelledForm = data_form.Form("cancel")
        element.addChild(cancelledForm.toElement())
        form = data_form.findForm(element, "myns")
        self.assertEqual("cancel", form.formType)

    def test_otherFormType(self):
        """
        Forms with other FORM_TYPEs are ignored.
        """
        element = domish.Element((None, "test"))
        otherForm = data_form.Form("submit", formNamespace="otherns")
        element.addChild(otherForm.toElement())
        form = data_form.findForm(element, "myns")
        self.assertIdentical(None, form)

    def test_otherFormTypeCancel(self):
        """
        Cancelled forms with another FORM_TYPE are ignored.
        """
        element = domish.Element((None, "test"))
        cancelledForm = data_form.Form("cancel", formNamespace="otherns")
        element.addChild(cancelledForm.toElement())
        form = data_form.findForm(element, "myns")
        self.assertIdentical(None, form)

    def test_noElement(self):
        """
        When None is passed as element, None is returned.
        """
        element = None
        form = data_form.findForm(element, "myns")
        self.assertIdentical(None, form)

    def test_noForm(self):
        """
        When no child element is a form, None is returned.
        """
        element = domish.Element((None, "test"))
        form = data_form.findForm(element, "myns")
        self.assertIdentical(None, form)

    def test_typeCheckNoFieldDefs(self):
        """
        If there are no field defs, an empty dictionary is assumed.
        """
        checked = []
        form = data_form.Form("submit")
        form.addField(
            data_form.Field("list-single", var="pubsub#description", value="a node")
        )
        field = form.fields["pubsub#description"]
        field.typeCheck = lambda: checked.append(None)
        form.typeCheck()

        self.assertIn("pubsub#description", form.fields)
        self.assertEqual([None], checked)



# -*- test-case-name: wokkel.test.test_server -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
XMPP Server-to-Server protocol.

This module implements several aspects of XMPP server-to-server communications
as described in XMPP Core (RFC 3920). Refer to that document for the meaning
of the used terminology.
"""

from __future__ import division, absolute_import

import binascii
from hashlib import sha256
import hmac

from zope.interface import implementer

from twisted.internet import defer, reactor
from twisted.names.srvconnect import SRVConnector
from twisted.python import log, randbytes
from twisted.words.protocols.jabber import error, ijabber, jid, xmlstream
from twisted.words.xish import domish

from wokkel.generic import DeferredXmlStreamFactory, XmlPipe

NS_DIALBACK = "jabber:server:dialback"


def generateKey(secret, receivingServer, originatingServer, streamID):
    """
    Generate a dialback key for server-to-server XMPP Streams.

    The dialback key is generated using the algorithm described in
    U{XEP-0185<http://xmpp.org/extensions/xep-0185.html>}. The used
    terminology for the parameters is described in RFC-3920.

    @param secret: the shared secret known to the Originating Server and
                   Authoritive Server.
    @type secret: L{str}
    @param receivingServer: the Receiving Server host name.
    @type receivingServer: L{str}
    @param originatingServer: the Originating Server host name.
    @type originatingServer: L{str}
    @param streamID: the Stream ID as generated by the Receiving Server.
    @type streamID: L{str}
    @return: hexadecimal digest of the generated key.
    @type: L{str}
    """

    hashObject = sha256()
    hashObject.update(secret.encode("ascii"))
    hashedSecret = hashObject.hexdigest()
    message = " ".join([receivingServer, originatingServer, streamID])
    hash = hmac.HMAC(
        hashedSecret.encode("ascii"), message.encode("ascii"), digestmod=sha256
    )
    return hash.hexdigest()


def trapStreamError(xs, observer):
    """
    Trap stream errors.

    This wraps an observer to catch exceptions. In case of a
    L{error.StreamError}, it is send over the given XML stream. All other
    exceptions yield a C{'internal-server-error'} stream error, that is
    sent over the stream, while the exception is logged.

    @return: Wrapped observer
    """

    def wrappedObserver(element):
        try:
            observer(element)
        except error.StreamError as exc:
            xs.sendStreamError(exc)
        except:
            log.err()
            exc = error.StreamError("internal-server-error")
            xs.sendStreamError(exc)

    return wrappedObserver


class XMPPServerConnector(SRVConnector):
    def __init__(self, reactor, domain, factory):
        SRVConnector.__init__(self, reactor, "xmpp-server", domain, factory)

    def pickServer(self):
        host, port = SRVConnector.pickServer(self)

        if not self.servers and not self.orderedServers:
            # no SRV record, fall back..
            port = 5269

        return host, port


class DialbackFailed(Exception):
    pass


@implementer(ijabber.IInitiatingInitializer)
class OriginatingDialbackInitializer(object):
    """
    Server Dialback Initializer for the Orginating Server.
    """

    _deferred = None

    def __init__(self, xs, thisHost, otherHost, secret):
        self.xmlstream = xs
        self.thisHost = thisHost
        self.otherHost = otherHost
        self.secret = secret

    def initialize(self):
        self._deferred = defer.Deferred()
        self.xmlstream.addObserver(xmlstream.STREAM_ERROR_EVENT, self.onStreamError)
        self.xmlstream.addObserver("/result[@xmlns='%s']" % NS_DIALBACK, self.onResult)

        key = generateKey(self.secret, self.otherHost, self.thisHost, self.xmlstream.sid)

        result = domish.Element((NS_DIALBACK, "result"))
        result["from"] = self.thisHost
        result["to"] = self.otherHost
        result.addContent(key)

        self.xmlstream.send(result)

        return self._deferred

    def onResult(self, result):
        self.xmlstream.removeObserver(xmlstream.STREAM_ERROR_EVENT, self.onStreamError)
        if result["type"] == "valid":
            self.xmlstream.otherEntity = jid.internJID(self.otherHost)
            self._deferred.callback(None)
        else:
            self._deferred.errback(DialbackFailed())

    def onStreamError(self, failure):
        self.xmlstream.removeObserver("/result[@xmlns='%s']" % NS_DIALBACK, self.onResult)
        self._deferred.errback(failure)


@implementer(ijabber.IInitiatingInitializer)
class ReceivingDialbackInitializer(object):
    """
    Server Dialback Initializer for the Receiving Server.
    """

    _deferred = None

    def __init__(self, xs, thisHost, otherHost, originalStreamID, key):
        self.xmlstream = xs
        self.thisHost = thisHost
        self.otherHost = otherHost
        self.originalStreamID = originalStreamID
        self.key = key

    def initialize(self):
        self._deferred = defer.Deferred()
        self.xmlstream.addObserver(xmlstream.STREAM_ERROR_EVENT, self.onStreamError)
        self.xmlstream.addObserver("/verify[@xmlns='%s']" % NS_DIALBACK, self.onVerify)

        verify = domish.Element((NS_DIALBACK, "verify"))
        verify["from"] = self.thisHost
        verify["to"] = self.otherHost
        verify["id"] = self.originalStreamID
        verify.addContent(self.key)

        self.xmlstream.send(verify)
        return self._deferred

    def onVerify(self, verify):
        self.xmlstream.removeObserver(xmlstream.STREAM_ERROR_EVENT, self.onStreamError)
        if verify["id"] != self.originalStreamID:
            self.xmlstream.sendStreamError(error.StreamError("invalid-id"))
            self._deferred.errback(DialbackFailed())
        elif verify["to"] != self.thisHost:
            self.xmlstream.sendStreamError(error.StreamError("host-unknown"))
            self._deferred.errback(DialbackFailed())
        elif verify["from"] != self.otherHost:
            self.xmlstream.sendStreamError(error.StreamError("invalid-from"))
            self._deferred.errback(DialbackFailed())
        elif verify["type"] == "valid":
            self._deferred.callback(None)
        else:
            self._deferred.errback(DialbackFailed())

    def onStreamError(self, failure):
        self.xmlstream.removeObserver("/verify[@xmlns='%s']" % NS_DIALBACK, self.onVerify)
        self._deferred.errback(failure)


class XMPPServerConnectAuthenticator(xmlstream.ConnectAuthenticator):
    """
    Authenticator for an outgoing XMPP server-to-server connection.

    This authenticator connects to C{otherHost} (the Receiving Server) and then
    initiates dialback as C{thisHost} (the Originating Server) using
    L{OriginatingDialbackInitializer}.

    @ivar thisHost: The domain this server connects from (the Originating
                    Server) .
    @ivar otherHost: The domain of the server this server connects to (the
                     Receiving Server).
    @ivar secret: The shared secret that is used for verifying the validity
                  of this new connection.
    """

    namespace = "jabber:server"

    def __init__(self, thisHost, otherHost, secret):
        self.thisHost = thisHost
        self.otherHost = otherHost
        self.secret = secret
        xmlstream.ConnectAuthenticator.__init__(self, otherHost)

    def connectionMade(self):
        self.xmlstream.thisEntity = jid.internJID(self.thisHost)
        self.xmlstream.prefixes = {xmlstream.NS_STREAMS: "stream", NS_DIALBACK: "db"}
        xmlstream.ConnectAuthenticator.connectionMade(self)

    def associateWithStream(self, xs):
        xmlstream.ConnectAuthenticator.associateWithStream(self, xs)
        init = OriginatingDialbackInitializer(
            xs, self.thisHost, self.otherHost, self.secret
        )
        xs.initializers = [init]


class XMPPServerVerifyAuthenticator(xmlstream.ConnectAuthenticator):
    """
    Authenticator for an outgoing connection to verify an incoming connection.

    This authenticator connects to C{otherHost} (the Authoritative Server) and
    then initiates dialback as C{thisHost} (the Receiving Server) using
    L{ReceivingDialbackInitializer}.

    @ivar thisHost: The domain this server connects from (the Receiving
                    Server) .
    @ivar otherHost: The domain of the server this server connects to (the
                     Authoritative Server).
    @ivar originalStreamID: The stream ID of the incoming connection that is
                            being verified.
    @ivar key: The key provided by the Receving Server to be verified.
    """

    namespace = "jabber:server"

    def __init__(self, thisHost, otherHost, originalStreamID, key):
        self.thisHost = thisHost
        self.otherHost = otherHost
        self.originalStreamID = originalStreamID
        self.key = key
        xmlstream.ConnectAuthenticator.__init__(self, otherHost)

    def connectionMade(self):
        self.xmlstream.thisEntity = jid.internJID(self.thisHost)
        self.xmlstream.prefixes = {xmlstream.NS_STREAMS: "stream", NS_DIALBACK: "db"}
        xmlstream.ConnectAuthenticator.connectionMade(self)

    def associateWithStream(self, xs):
        xmlstream.ConnectAuthenticator.associateWithStream(self, xs)
        init = ReceivingDialbackInitializer(
            xs, self.thisHost, self.otherHost, self.originalStreamID, self.key
        )
        xs.initializers = [init]


class XMPPServerListenAuthenticator(xmlstream.ListenAuthenticator):
    """
    Authenticator for an incoming XMPP server-to-server connection.

    This authenticator handles two types of incoming connections. Regular
    server-to-server connections are from the Originating Server to the
    Receiving Server, where this server is the Receiving Server. These
    connections start out by receiving a dialback key, verifying the
    key with the Authoritative Server, and then accept normal XMPP stanzas.

    The other type of connections is from a Receiving Server to an
    Authoritative Server, where this server acts as the Authoritative Server.
    These connections are used to verify the validity of an outgoing connection
    from this server. In this case, this server receives a verification
    request, checks the key and then returns the result.

    @ivar service: The service that keeps the list of domains we accept
                   connections for.
    """

    namespace = "jabber:server"

    def __init__(self, service):
        xmlstream.ListenAuthenticator.__init__(self)
        self.service = service

    def streamStarted(self, rootElement):
        xmlstream.ListenAuthenticator.streamStarted(self, rootElement)

        if self.xmlstream.thisEntity:
            targetDomain = self.xmlstream.thisEntity.host
        else:
            targetDomain = self.service.defaultDomain

        def prepareStream(domain):
            self.xmlstream.namespace = self.namespace
            self.xmlstream.prefixes = {xmlstream.NS_STREAMS: "stream", NS_DIALBACK: "db"}
            if domain:
                self.xmlstream.thisEntity = jid.internJID(domain)

        try:
            if (
                xmlstream.NS_STREAMS != rootElement.uri
                or self.namespace != self.xmlstream.namespace
                or ("db", NS_DIALBACK) not in rootElement.localPrefixes.items()
            ):
                raise error.StreamError("invalid-namespace")

            if targetDomain and targetDomain not in self.service.domains:
                raise error.StreamError("host-unknown")
        except error.StreamError as exc:
            prepareStream(self.service.defaultDomain)
            self.xmlstream.sendStreamError(exc)
            return

        self.xmlstream.addObserver(
            "//verify[@xmlns='%s']" % NS_DIALBACK,
            trapStreamError(self.xmlstream, self.onVerify),
        )
        self.xmlstream.addObserver("//result[@xmlns='%s']" % NS_DIALBACK, self.onResult)

        prepareStream(targetDomain)
        self.xmlstream.sendHeader()

        if self.xmlstream.version >= (1, 0):
            features = domish.Element((xmlstream.NS_STREAMS, "features"))
            self.xmlstream.send(features)

    def onVerify(self, verify):
        try:
            receivingServer = jid.JID(verify["from"]).host
            originatingServer = jid.JID(verify["to"]).host
        except (KeyError, jid.InvalidFormat):
            raise error.StreamError("improper-addressing")

        if originatingServer not in self.service.domains:
            raise error.StreamError("host-unknown")

        if (
            self.xmlstream.otherEntity
            and receivingServer != self.xmlstream.otherEntity.host
        ):
            raise error.StreamError("invalid-from")

        streamID = verify.getAttribute("id", "")
        key = str(verify)

        calculatedKey = generateKey(
            self.service.secret, receivingServer, originatingServer, streamID
        )
        validity = (key == calculatedKey) and "valid" or "invalid"

        reply = domish.Element((NS_DIALBACK, "verify"))
        reply["from"] = originatingServer
        reply["to"] = receivingServer
        reply["id"] = streamID
        reply["type"] = validity
        self.xmlstream.send(reply)

    def onResult(self, result):
        def reply(validity):
            reply = domish.Element((NS_DIALBACK, "result"))
            reply["from"] = result["to"]
            reply["to"] = result["from"]
            reply["type"] = validity
            self.xmlstream.send(reply)

        def valid(xs):
            reply("valid")
            if not self.xmlstream.thisEntity:
                self.xmlstream.thisEntity = jid.internJID(receivingServer)
            self.xmlstream.otherEntity = jid.internJID(originatingServer)
            self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)

        def invalid(failure):
            log.err(failure)
            reply("invalid")

        receivingServer = result["to"]
        originatingServer = result["from"]
        key = str(result)

        d = self.service.validateConnection(
            receivingServer, originatingServer, self.xmlstream.sid, key
        )
        d.addCallbacks(valid, invalid)
        return d


class DeferredS2SClientFactory(DeferredXmlStreamFactory):
    """
    Deferred firing factory for initiating XMPP server-to-server connection.

    The deferred has its callbacks called upon succesful authentication with
    the other server. In case of failed authentication or connection, the
    deferred will have its errbacks called instead.
    """

    logTraffic = False

    def __init__(self, authenticator):
        DeferredXmlStreamFactory.__init__(self, authenticator)

        self.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self.onConnectionMade)

        self.serial = 0

    def onConnectionMade(self, xs):
        xs.serial = self.serial
        self.serial += 1

        def logDataIn(buf):
            log.msg("RECV (%d): %r" % (xs.serial, buf))

        def logDataOut(buf):
            log.msg("SEND (%d): %r" % (xs.serial, buf))

        if self.logTraffic:
            xs.rawDataInFn = logDataIn
            xs.rawDataOutFn = logDataOut


def initiateS2S(factory):
    domain = factory.authenticator.otherHost.encode("idna")
    c = XMPPServerConnector(reactor, domain, factory)
    c.connect()
    return factory.deferred


class XMPPS2SServerFactory(xmlstream.XmlStreamServerFactory):
    """
    XMPP Server-to-Server Server factory.

    This factory accepts XMPP server-to-server connections.
    """

    logTraffic = False

    def __init__(self, service):
        self.service = service

        def authenticatorFactory():
            return XMPPServerListenAuthenticator(service)

        xmlstream.XmlStreamServerFactory.__init__(self, authenticatorFactory)
        self.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self.onConnectionMade)
        self.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self.onAuthenticated)

        self.serial = 0

    def onConnectionMade(self, xs):
        """
        Called when a server-to-server connection was made.

        This enables traffic debugging on incoming streams.
        """
        xs.serial = self.serial
        self.serial += 1

        def logDataIn(buf):
            log.msg("RECV (%d): %r" % (xs.serial, buf))

        def logDataOut(buf):
            log.msg("SEND (%d): %r" % (xs.serial, buf))

        if self.logTraffic:
            xs.rawDataInFn = logDataIn
            xs.rawDataOutFn = logDataOut

        xs.addObserver(xmlstream.STREAM_ERROR_EVENT, self.onError)

    def onAuthenticated(self, xs):
        thisHost = xs.thisEntity.host
        otherHost = xs.otherEntity.host

        log.msg(
            "Incoming connection %d from %r to %r established"
            % (xs.serial, otherHost, thisHost)
        )

        xs.addObserver(xmlstream.STREAM_END_EVENT, self.onConnectionLost, 0, xs)
        xs.addObserver("/*", self.onElement, 0, xs)

    def onConnectionLost(self, xs, reason):
        thisHost = xs.thisEntity.host
        otherHost = xs.otherEntity.host

        log.msg(
            "Incoming connection %d from %r to %r disconnected"
            % (xs.serial, otherHost, thisHost)
        )

    def onError(self, reason):
        log.err(reason, "Stream Error")

    def onElement(self, xs, element):
        """
        Called when an element was received from one of the connected streams.

        """
        if element.handled:
            return
        else:
            self.service.dispatch(xs, element)


class ServerService(object):
    """
    Service for managing XMPP server to server connections.
    """

    logTraffic = False

    def __init__(self, router, domain=None, secret=None):
        self.router = router

        self.defaultDomain = domain
        self.domains = set()
        if self.defaultDomain:
            self.domains.add(self.defaultDomain)

        if secret is not None:
            self.secret = secret
        else:
            self.secret = binascii.hexlify(randbytes.secureRandom(16))

        self._outgoingStreams = {}
        self._outgoingQueues = {}
        self._outgoingConnecting = set()
        self.serial = 0

        pipe = XmlPipe()
        self.xmlstream = pipe.source
        self.router.addRoute(None, pipe.sink)
        self.xmlstream.addObserver("/*", self.send)

    def outgoingInitialized(self, xs):
        thisHost = xs.thisEntity.host
        otherHost = xs.otherEntity.host

        log.msg(
            "Outgoing connection %d from %r to %r established"
            % (xs.serial, thisHost, otherHost)
        )

        self._outgoingStreams[thisHost, otherHost] = xs
        xs.addObserver(
            xmlstream.STREAM_END_EVENT, lambda _: self.outgoingDisconnected(xs)
        )

        if (thisHost, otherHost) in self._outgoingQueues:
            for element in self._outgoingQueues[thisHost, otherHost]:
                xs.send(element)
            del self._outgoingQueues[thisHost, otherHost]

    def outgoingDisconnected(self, xs):
        thisHost = xs.thisEntity.host
        otherHost = xs.otherEntity.host

        log.msg(
            "Outgoing connection %d from %r to %r disconnected"
            % (xs.serial, thisHost, otherHost)
        )

        del self._outgoingStreams[thisHost, otherHost]

    def initiateOutgoingStream(self, thisHost, otherHost):
        """
        Initiate an outgoing XMPP server-to-server connection.
        """

        def resetConnecting(_):
            self._outgoingConnecting.remove((thisHost, otherHost))

        if (thisHost, otherHost) in self._outgoingConnecting:
            return

        authenticator = XMPPServerConnectAuthenticator(thisHost, otherHost, self.secret)
        factory = DeferredS2SClientFactory(authenticator)
        factory.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self.outgoingInitialized)
        factory.logTraffic = self.logTraffic

        self._outgoingConnecting.add((thisHost, otherHost))

        d = initiateS2S(factory)
        d.addBoth(resetConnecting)
        return d

    def validateConnection(self, thisHost, otherHost, sid, key):
        """
        Validate an incoming XMPP server-to-server connection.
        """

        def connected(xs):
            # Set up stream for immediate disconnection.
            def disconnect(_):
                xs.transport.loseConnection()

            xs.addObserver(xmlstream.STREAM_AUTHD_EVENT, disconnect)
            xs.addObserver(xmlstream.INIT_FAILED_EVENT, disconnect)

        authenticator = XMPPServerVerifyAuthenticator(thisHost, otherHost, sid, key)
        factory = DeferredS2SClientFactory(authenticator)
        factory.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, connected)
        factory.logTraffic = self.logTraffic

        d = initiateS2S(factory)
        return d

    def send(self, stanza):
        """
        Send stanza to the proper XML Stream.

        This uses addressing embedded in the stanza to find the correct stream
        to forward the stanza to.
        """

        otherHost = jid.internJID(stanza["to"]).host
        thisHost = jid.internJID(stanza["from"]).host

        if (thisHost, otherHost) not in self._outgoingStreams:
            # There is no connection with the destination (yet). Cache the
            # outgoing stanza until the connection has been established.
            # XXX: If the connection cannot be established, the queue should
            #      be emptied at some point.
            if (thisHost, otherHost) not in self._outgoingQueues:
                self._outgoingQueues[(thisHost, otherHost)] = []
            self._outgoingQueues[(thisHost, otherHost)].append(stanza)
            self.initiateOutgoingStream(thisHost, otherHost)
        else:
            self._outgoingStreams[(thisHost, otherHost)].send(stanza)

    def dispatch(self, xs, stanza):
        """
        Send on element to be routed within the server.
        """
        stanzaFrom = stanza.getAttribute("from")
        stanzaTo = stanza.getAttribute("to")

        if not stanzaFrom or not stanzaTo:
            xs.sendStreamError(error.StreamError("improper-addressing"))
        else:
            try:
                sender = jid.internJID(stanzaFrom)
                jid.internJID(stanzaTo)
            except jid.InvalidFormat:
                log.msg("Dropping error stanza with malformed JID")

            if sender.host != xs.otherEntity.host:
                xs.sendStreamError(error.StreamError("invalid-from"))
            else:
                self.xmlstream.send(stanza)



# -*- test-case-name: wokkel.test.test_muc -*-
#
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
XMPP Multi-User Chat protocol.

This protocol is specified in
U{XEP-0045<http://xmpp.org/extensions/xep-0045.html>}.
"""

from __future__ import absolute_import, division

from constantly import ValueConstant, Values
from dateutil.tz import tzutc
from twisted.internet import defer
from twisted.words.protocols.jabber import error, jid, xmlstream
from twisted.words.xish import domish
from zope.interface import implementer

from wokkel import data_form, generic, iwokkel, xmppim
from wokkel.delay import Delay, DelayMixin
from wokkel.iwokkel import IMUCClient
from wokkel.subprotocols import XMPPHandler

# Multi User Chat namespaces
NS_MUC = "http://jabber.org/protocol/muc"
NS_MUC_USER = NS_MUC + "#user"
NS_MUC_ADMIN = NS_MUC + "#admin"
NS_MUC_OWNER = NS_MUC + "#owner"
NS_MUC_ROOMINFO = NS_MUC + "#roominfo"
NS_MUC_CONFIG = NS_MUC + "#roomconfig"
NS_MUC_REQUEST = NS_MUC + "#request"
NS_MUC_REGISTER = NS_MUC + "#register"

NS_REGISTER = "jabber:iq:register"

MESSAGE = "/message"
PRESENCE = "/presence"

GROUPCHAT = MESSAGE + '[@type="groupchat"]'

DEFER_TIMEOUT = 30  # basic timeout is 30 seconds


class STATUS_CODE(Values):
    REALJID_PUBLIC = ValueConstant(100)
    AFFILIATION_CHANGED = ValueConstant(101)
    UNAVAILABLE_SHOWN = ValueConstant(102)
    UNAVAILABLE_NOT_SHOWN = ValueConstant(103)
    CONFIGURATION_CHANGED = ValueConstant(104)
    SELF_PRESENCE = ValueConstant(110)
    LOGGING_ENABLED = ValueConstant(170)
    LOGGING_DISABLED = ValueConstant(171)
    NON_ANONYMOUS = ValueConstant(172)
    SEMI_ANONYMOUS = ValueConstant(173)
    FULLY_ANONYMOUS = ValueConstant(174)
    ROOM_CREATED = ValueConstant(201)
    NICK_ASSIGNED = ValueConstant(210)
    BANNED = ValueConstant(301)
    NEW_NICK = ValueConstant(303)
    KICKED = ValueConstant(307)
    REMOVED_AFFILIATION = ValueConstant(321)
    REMOVED_MEMBERSHIP = ValueConstant(322)
    REMOVED_SHUTDOWN = ValueConstant(332)


@implementer(iwokkel.IMUCStatuses)
class Statuses(set):
    """
    Container of MUC status conditions.

    This is currently implemented as a set of constant values from
    L{STATUS_CODE}. Instances of this class provide L{IMUCStatuses}, that
    defines the supported operations. Even though this class currently derives
    from L{set}, future versions might not. This provides an upgrade path to
    cater for extensible status conditions, as defined in
    U{XEP-0306<http://xmpp.org/extensions/xep-0306.html>}.
    """


class _FormRequest(generic.Request):
    """
    Base class for form exchange requests.
    """

    requestNamespace = None
    formNamespace = None

    def __init__(self, recipient, sender=None, options=None):
        if options is None:
            stanzaType = "get"
        else:
            stanzaType = "set"

        generic.Request.__init__(self, recipient, sender, stanzaType)
        self.options = options

    def toElement(self):
        element = generic.Request.toElement(self)

        query = element.addElement((self.requestNamespace, "query"))
        if self.options is None:
            # This is a request for the configuration form.
            form = None
        elif self.options is False:
            form = data_form.Form(formType="cancel")
        else:
            form = data_form.Form(formType="submit", formNamespace=self.formNamespace)
            form.makeFields(self.options)

        if form is not None:
            query.addChild(form.toElement())

        return element


class ConfigureRequest(_FormRequest):
    """
    Configure MUC room request.

    http://xmpp.org/extensions/xep-0045.html#roomconfig
    """

    requestNamespace = NS_MUC_OWNER
    formNamespace = NS_MUC_CONFIG


class RegisterRequest(_FormRequest):
    """
    Register request.

    http://xmpp.org/extensions/xep-0045.html#register
    """

    requestNamespace = NS_REGISTER
    formNamespace = NS_MUC_REGISTER


class AdminItem(object):
    """
    Item representing role and/or affiliation for admin request.
    """

    def __init__(self, affiliation=None, role=None, entity=None, nick=None, reason=None):
        self.affiliation = affiliation
        self.role = role
        self.entity = entity
        self.nick = nick
        self.reason = reason

    def toElement(self):
        element = domish.Element((NS_MUC_ADMIN, "item"))

        if self.entity:
            element["jid"] = self.entity.full()

        if self.nick:
            element["nick"] = self.nick

        if self.affiliation:
            element["affiliation"] = self.affiliation

        if self.role:
            element["role"] = self.role

        if self.reason:
            element.addElement("reason", content=self.reason)

        return element

    @classmethod
    def fromElement(Class, element):
        item = Class()

        if element.hasAttribute("jid"):
            item.entity = jid.JID(element["jid"])

        item.nick = element.getAttribute("nick")
        item.affiliation = element.getAttribute("affiliation")
        item.role = element.getAttribute("role")

        for child in element.elements(NS_MUC_ADMIN, "reason"):
            item.reason = str(child)

        return item


class AdminStanza(generic.Request):
    """
    An admin request or response.
    """

    childParsers = {(NS_MUC_ADMIN, "query"): "_childParser_query"}

    def toElement(self):
        element = generic.Request.toElement(self)
        element.addElement((NS_MUC_ADMIN, "query"))

        if self.items:
            for item in self.items:
                element.query.addChild(item.toElement())

        return element

    def _childParser_query(self, element):
        self.items = []
        for child in element.elements(NS_MUC_ADMIN, "item"):
            self.items.append(AdminItem.fromElement(child))


class DestructionRequest(generic.Request):
    """
    Room destruction request.

    @param reason: Optional reason for the destruction of this room.
    @type reason: L{str}.

    @param alternate: Optional room JID of an alternate venue.
    @type alternate: L{JID<twisted.words.protocols.jabber.jid.JID>}

    @param password: Optional password for entering the alternate venue.
    @type password: L{str}
    """

    stanzaType = "set"

    def __init__(
        self, recipient, sender=None, reason=None, alternate=None, password=None
    ):
        generic.Request.__init__(self, recipient, sender)
        self.reason = reason
        self.alternate = alternate
        self.password = password

    def toElement(self):
        element = generic.Request.toElement(self)
        element.addElement((NS_MUC_OWNER, "query"))
        element.query.addElement("destroy")

        if self.alternate:
            element.query.destroy["jid"] = self.alternate.full()

            if self.password:
                element.query.destroy.addElement("password", content=self.password)

        if self.reason:
            element.query.destroy.addElement("reason", content=self.reason)

        return element


class GroupChat(xmppim.Message, DelayMixin):
    """
    A groupchat message.
    """

    stanzaType = "groupchat"

    def toElement(self, legacyDelay=False):
        """
        Render into a domish Element.

        @param legacyDelay: If L{True} send the delayed delivery information
        in legacy format.
        """
        element = xmppim.Message.toElement(self)

        if self.delay:
            element.addChild(self.delay.toElement(legacy=legacyDelay))

        return element


class PrivateChat(xmppim.Message):
    """
    A chat message.
    """

    stanzaType = "chat"


class InviteMessage(xmppim.Message):

    def __init__(self, recipient=None, sender=None, invitee=None, reason=None):
        xmppim.Message.__init__(self, recipient, sender)
        self.invitee = invitee
        self.reason = reason

    def toElement(self):
        element = xmppim.Message.toElement(self)

        child = element.addElement((NS_MUC_USER, "x"))
        child.addElement("invite")
        child.invite["to"] = self.invitee.full()

        if self.reason:
            child.invite.addElement("reason", content=self.reason)

        return element


class HistoryOptions(object):
    """
    A history configuration object.

    @ivar maxchars: Limit the total number of characters in the history to "X"
        (where the character count is the characters of the complete XML
        stanzas, not only their XML character data).
    @type maxchars: L{int}

    @ivar maxstanzas: Limit the total number of messages in the history to "X".
    @type mazstanzas: L{int}

    @ivar seconds: Send only the messages received in the last "X" seconds.
    @type seconds: L{int}

    @ivar since: Send only the messages received since the datetime specified.
        Note that this must be an offset-aware instance.
    @type since: L{datetime.datetime}
    """

    attributes = ["maxChars", "maxStanzas", "seconds", "since"]

    def __init__(self, maxChars=None, maxStanzas=None, seconds=None, since=None):
        self.maxChars = maxChars
        self.maxStanzas = maxStanzas
        self.seconds = seconds
        self.since = since

    def toElement(self):
        """
        Returns a L{domish.Element} representing the history options.
        """
        element = domish.Element((NS_MUC, "history"))

        for key in self.attributes:
            value = getattr(self, key, None)
            if value is not None:
                if key == "since":
                    stamp = value.astimezone(tzutc())
                    element[key] = stamp.strftime("%Y-%m-%dT%H:%M:%SZ")
                else:
                    element[key.lower()] = str(value)

        return element


class BasicPresence(xmppim.AvailabilityPresence):
    """
    Availability presence sent from MUC client to service.

    @type history: L{HistoryOptions}
    """

    history = None
    password = None

    def toElement(self):
        element = xmppim.AvailabilityPresence.toElement(self)

        muc = element.addElement((NS_MUC, "x"))
        if self.password:
            muc.addElement("password", content=self.password)
        if self.history:
            muc.addChild(self.history.toElement())

        return element


class UserPresence(xmppim.AvailabilityPresence):
    """
    Availability presence sent from MUC service to client.

    @ivar affiliation: Affiliation of the entity to the room.
    @type affiliation: L{str}

    @ivar role: Role of the entity in the room.
    @type role: L{str}

    @ivar entity: The real JID of the entity this presence is from.
    @type entity: L{JID<twisted.words.protocols.jabber.jid.JID>}

    @ivar mucStatuses: Set of one or more status codes from L{STATUS_CODE}.
        See L{Statuses} for usage notes.
    @type mucStatuses: L{Statuses}

    @ivar nick: The nick name of the entity in the room.
    @type nick: L{str}
    """

    affiliation = None
    role = None
    entity = None
    nick = None

    mucStatuses = None

    childParsers = {(NS_MUC_USER, "x"): "_childParser_mucUser"}

    def __init__(self, *args, **kwargs):
        self.mucStatuses = Statuses()
        xmppim.AvailabilityPresence.__init__(self, *args, **kwargs)

    def _childParser_mucUser(self, element):
        """
        Parse the MUC user extension element.
        """
        for child in element.elements():
            if child.uri != NS_MUC_USER:
                continue

            elif child.name == "status":
                try:
                    value = int(child.getAttribute("code"))
                    statusCode = STATUS_CODE.lookupByValue(value)
                except (TypeError, ValueError):
                    continue

                self.mucStatuses.add(statusCode)

            elif child.name == "item":
                if child.hasAttribute("jid"):
                    self.entity = jid.JID(child["jid"])

                self.nick = child.getAttribute("nick")
                self.affiliation = child.getAttribute("affiliation")
                self.role = child.getAttribute("role")

                for reason in child.elements(NS_MUC_ADMIN, "reason"):
                    self.reason = str(reason)

            # TODO: destroy


class VoiceRequest(xmppim.Message):
    """
    Voice request message.
    """

    def toElement(self):
        element = xmppim.Message.toElement(self)

        # build data form
        form = data_form.Form("submit", formNamespace=NS_MUC_REQUEST)
        form.addField(
            data_form.Field(var="muc#role", value="participant", label="Requested role")
        )
        element.addChild(form.toElement())

        return element


class MUCClientProtocol(xmppim.BasePresenceProtocol):
    """
    Multi-User Chat client protocol.
    """

    timeout = None

    presenceTypeParserMap = {
        "error": generic.ErrorStanza,
        "available": UserPresence,
        "unavailable": UserPresence,
    }

    def __init__(self, reactor=None):
        XMPPHandler.__init__(self)

        if reactor:
            self._reactor = reactor
        else:
            from twisted.internet import reactor

            self._reactor = reactor

    def connectionInitialized(self):
        """
        Called when the XML stream has been initialized.

        It initializes several XPath events to handle MUC stanzas that come
        in.
        """
        xmppim.BasePresenceProtocol.connectionInitialized(self)
        self.xmlstream.addObserver(GROUPCHAT, self._onGroupChat)
        self._roomOccupantMap = {}

    def _onGroupChat(self, element):
        """
        A group chat message has been received from a MUC room.

        There are a few event methods that may get called here.
        L{receivedGroupChat}, L{receivedSubject} or L{receivedHistory}.
        """
        message = GroupChat.fromElement(element)
        self.groupChatReceived(message)

    def groupChatReceived(self, message):
        """
        Called when a groupchat message was received.

        This method is called with a parsed representation of a received
        groupchat message and can be overridden for further processing.

        For regular groupchat message, the C{body} attribute contains the
        message body. Conversation history sent by the room upon joining, will
        have the C{delay} attribute set, room subject changes the C{subject}
        attribute. See L{GroupChat} for details.

        @param message: Groupchat message.
        @type message: L{GroupChat}
        """
        pass

    def _sendDeferred(self, stanza):
        """
        Send presence stanza, adding a deferred with a timeout.

        @param stanza: The presence stanza to send over the wire.
        @type stanza: L{generic.Stanza}

        @param timeout: The number of seconds to wait before the deferred is
            timed out.
        @type timeout: L{int}

        The deferred object L{defer.Deferred} is returned.
        """

        def onResponse(element):
            if element.getAttribute("type") == "error":
                d.errback(error.exceptionFromStanza(element))
            else:
                d.callback(UserPresence.fromElement(element))

        def onTimeout():
            d.errback(xmlstream.TimeoutError("Timeout waiting for response."))

        def cancelTimeout(result):
            if call.active():
                call.cancel()

            return result

        def recordOccupant(presence):
            occupantJID = presence.sender
            roomJID = occupantJID.userhostJID()
            self._roomOccupantMap[roomJID] = occupantJID
            return presence

        call = self._reactor.callLater(DEFER_TIMEOUT, onTimeout)

        d = defer.Deferred()
        d.addBoth(cancelTimeout)
        d.addCallback(recordOccupant)

        query = "/presence[@from='%s' or (@from='%s' and @type='error')]" % (
            stanza.recipient.full(),
            stanza.recipient.userhost(),
        )
        self.xmlstream.addOnetimeObserver(query, onResponse, priority=-1)
        self.xmlstream.send(stanza.toElement())
        return d

    def join(self, roomJID, nick, historyOptions=None, password=None):
        """
        Join a MUC room by sending presence to it.

        @param roomJID: The JID of the room the entity is joining.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The nick name for the entitity joining the room.
        @type nick: L{str}

        @param historyOptions: Options for conversation history sent by the
            room upon joining.
        @type historyOptions: L{HistoryOptions}

        @param password: Optional password for the room.
        @type password: L{str}

        @return: A deferred that fires when the entity is in the room or an
                 error has occurred.
        """
        occupantJID = jid.JID(tuple=(roomJID.user, roomJID.host, nick))

        presence = BasicPresence(recipient=occupantJID)
        if password:
            presence.password = password
        if historyOptions:
            presence.history = historyOptions

        return self._sendDeferred(presence)

    def nick(self, roomJID, nick):
        """
        Change an entity's nick name in a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#changenick

        @param roomJID: The JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The new nick name within the room.
        @type nick: L{str}
        """
        occupantJID = jid.JID(tuple=(roomJID.user, roomJID.host, nick))
        presence = BasicPresence(recipient=occupantJID)
        return self._sendDeferred(presence)

    def status(self, roomJID, show=None, status=None):
        """
        Change user status.

        See: http://xmpp.org/extensions/xep-0045.html#changepres

        @param roomJID: The Room JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param show: The availability of the entity. Common values are xa,
            available, etc
        @type show: L{str}

        @param status: The current status of the entity.
        @type status: L{str}
        """
        occupantJID = self._roomOccupantMap[roomJID]
        presence = BasicPresence(recipient=occupantJID, show=show, status=status)
        return self._sendDeferred(presence)

    def leave(self, roomJID):
        """
        Leave a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#exit

        @param roomJID: The JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        occupantJID = self._roomOccupantMap[roomJID]
        presence = xmppim.AvailabilityPresence(recipient=occupantJID, available=False)

        return self._sendDeferred(presence)

    def groupChat(self, roomJID, body):
        """
        Send a groupchat message.
        """
        message = GroupChat(recipient=roomJID, body=body)
        self.send(message.toElement())

    def chat(self, occupantJID, body):
        """
        Send a private chat message to a user in a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#privatemessage

        @param occupantJID: The Room JID of the other user.
        @type occupantJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        message = PrivateChat(recipient=occupantJID, body=body)
        self.send(message.toElement())

    def subject(self, roomJID, subject):
        """
        Change the subject of a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#subject-mod

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param subject: The subject you want to set.
        @type subject: L{str}
        """
        message = GroupChat(roomJID.userhostJID(), subject=subject)
        self.send(message.toElement())

    def invite(self, roomJID, invitee, reason=None):
        """
        Invite a xmpp entity to a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#invite

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param invitee: The entity that is being invited.
        @type invitee: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param reason: The reason for the invite.
        @type reason: L{str}
        """
        message = InviteMessage(recipient=roomJID, invitee=invitee, reason=reason)
        self.send(message.toElement())

    def getRegisterForm(self, roomJID):
        """
        Grab the registration form for a MUC room.

        @param room: The room jabber/xmpp entity id for the requested
            registration form.
        @type room: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """

        def cb(response):
            form = data_form.findForm(response.query, NS_MUC_REGISTER)
            return form

        request = RegisterRequest(recipient=roomJID, options=None)
        d = self.request(request)
        d.addCallback(cb)
        return d

    def register(self, roomJID, options):
        """
        Send a request to register for a room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param options: A mapping of field names to values, or L{None} to
            cancel.
        @type options: L{dict}
        """
        if options is None:
            options = False
        request = RegisterRequest(recipient=roomJID, options=options)
        return self.request(request)

    def voice(self, roomJID):
        """
        Request voice for a moderated room.

        @param roomJID: The room jabber/xmpp entity id.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        message = VoiceRequest(recipient=roomJID)
        self.xmlstream.send(message.toElement())

    def history(self, roomJID, messages):
        """
        Send history to create a MUC based on a one on one chat.

        See: http://xmpp.org/extensions/xep-0045.html#continue

        @param roomJID: The room jabber/xmpp entity id.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param messages: The history to send to the room as an ordered list of
                         message, represented by a dictionary with the keys
                         L{'stanza'}, holding the original stanza a
                         L{domish.Element}, and L{'timestamp'} with the
                         timestamp.
        @type messages: L{list} of L{domish.Element}
        """

        for message in messages:
            stanza = message["stanza"]
            stanza["type"] = "groupchat"

            delay = Delay(stamp=message["timestamp"])

            sender = stanza.getAttribute("from")
            if sender is not None:
                delay.sender = jid.JID(sender)

            stanza.addChild(delay.toElement())

            stanza["to"] = roomJID.userhost()
            if stanza.hasAttribute("from"):
                del stanza["from"]

            self.xmlstream.send(stanza)

    def getConfiguration(self, roomJID):
        """
        Grab the configuration from the room.

        This sends an iq request to the room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @return: A deferred that fires with the room's configuration form as
            a L{data_form.Form} or L{None} if there are no configuration
            options available.
        """

        def cb(response):
            form = data_form.findForm(response.query, NS_MUC_CONFIG)
            return form

        request = ConfigureRequest(recipient=roomJID, options=None)
        d = self.request(request)
        d.addCallback(cb)
        return d

    def configure(self, roomJID, options):
        """
        Configure a room.

        @param roomJID: The room to configure.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param options: A mapping of field names to values, or L{None} to
            cancel.
        @type options: L{dict}
        """
        if options is None:
            options = False
        request = ConfigureRequest(recipient=roomJID, options=options)
        return self.request(request)

    def _getAffiliationList(self, roomJID, affiliation):
        """
        Send a request for an affiliation list in a room.
        """

        def cb(response):
            stanza = AdminStanza.fromElement(response)
            return stanza.items

        request = AdminStanza(recipient=roomJID, stanzaType="get")
        request.items = [AdminItem(affiliation=affiliation)]
        d = self.request(request)
        d.addCallback(cb)
        return d

    def _getRoleList(self, roomJID, role):
        """
        Send a request for a role list in a room.
        """

        def cb(response):
            stanza = AdminStanza.fromElement(response)
            return stanza.items

        request = AdminStanza(recipient=roomJID, stanzaType="get")
        request.items = [AdminItem(role=role)]
        d = self.request(request)
        d.addCallback(cb)
        return d

    def getMemberList(self, roomJID):
        """
        Get the member list of a room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._getAffiliationList(roomJID, "member")

    def getAdminList(self, roomJID):
        """
        Get the admin list of a room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._getAffiliationList(roomJID, "admin")

    def getBanList(self, roomJID):
        """
        Get an outcast list from a room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._getAffiliationList(roomJID, "outcast")

    def getOwnerList(self, roomJID):
        """
        Get an owner list from a room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._getAffiliationList(roomJID, "owner")

    def getModeratorList(self, roomJID):
        """
        Get the moderator list of a room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        d = self._getRoleList(roomJID, "moderator")
        return d

    def _setAffiliation(self, roomJID, entity, affiliation, reason=None, sender=None):
        """
        Send a request to change an entity's affiliation to a MUC room.
        """
        request = AdminStanza(recipient=roomJID, sender=sender, stanzaType="set")
        item = AdminItem(entity=entity, affiliation=affiliation, reason=reason)
        request.items = [item]
        return self.request(request)

    def _setRole(self, roomJID, nick, role, reason=None, sender=None):
        """
        Send a request to change an occupant's role in a MUC room.
        """
        request = AdminStanza(recipient=roomJID, sender=sender, stanzaType="set")
        item = AdminItem(nick=nick, role=role, reason=reason)
        request.items = [item]
        return self.request(request)

    def modifyAffiliationList(self, roomJID, entities, affiliation, sender=None):
        """
        Modify an affiliation list.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param entities: The list of entities to change for a room.
        @type entities: L{list} of
            L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param affiliation: The affilation to the entities will acquire.
        @type affiliation: L{str}

        @param sender: The entity sending the request.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}

        """
        request = AdminStanza(recipient=roomJID, sender=sender, stanzaType="set")
        request.items = [
            AdminItem(entity=entity, affiliation=affiliation) for entity in entities
        ]

        return self.request(request)

    def grantVoice(self, roomJID, nick, reason=None, sender=None):
        """
        Grant voice to an entity.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The nick name for the user in this room.
        @type nick: L{str}

        @param reason: The reason for granting voice to the entity.
        @type reason: L{str}

        @param sender: The entity sending the request.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._setRole(
            roomJID, nick=nick, role="participant", reason=reason, sender=sender
        )

    def revokeVoice(self, roomJID, nick, reason=None, sender=None):
        """
        Revoke voice from a participant.

        This will disallow the entity to send messages to a moderated room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The nick name for the user in this room.
        @type nick: L{str}

        @param reason: The reason for revoking voice from the entity.
        @type reason: L{str}

        @param sender: The entity sending the request.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._setRole(
            roomJID, nick=nick, role="visitor", reason=reason, sender=sender
        )

    def grantModerator(self, roomJID, nick, reason=None, sender=None):
        """
        Grant moderator privileges to a MUC room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The nick name for the user in this room.
        @type nick: L{str}

        @param reason: The reason for granting moderation to the entity.
        @type reason: L{str}

        @param sender: The entity sending the request.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._setRole(
            roomJID, nick=nick, role="moderator", reason=reason, sender=sender
        )

    def ban(self, roomJID, entity, reason=None, sender=None):
        """
        Ban a user from a MUC room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param entity: The bare JID of the entity to be banned.
        @type entity: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param reason: The reason for banning the entity.
        @type reason: L{str}

        @param sender: The entity sending the request.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._setAffiliation(
            roomJID, entity, "outcast", reason=reason, sender=sender
        )

    def kick(self, roomJID, nick, reason=None, sender=None):
        """
        Kick a user from a MUC room.

        @param roomJID: The bare JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The occupant to be banned.
        @type nick: L{str}

        @param reason: The reason given for the kick.
        @type reason: L{str}

        @param sender: The entity sending the request.
        @type sender: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._setRole(roomJID, nick, "none", reason=reason, sender=sender)

    def destroy(self, roomJID, reason=None, alternate=None, password=None):
        """
        Destroy a room.

        @param roomJID: The JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param reason: The reason for the destruction of the room.
        @type reason: L{str}

        @param alternate: The JID of the room suggested as an alternate venue.
        @type alternate: L{JID<twisted.words.protocols.jabber.jid.JID>}

        """
        request = DestructionRequest(
            recipient=roomJID, reason=reason, alternate=alternate, password=password
        )

        return self.request(request)


class User(object):
    """
    A user/entity in a multi-user chat room.
    """

    def __init__(self, nick, entity=None):
        self.nick = nick
        self.entity = entity
        self.affiliation = "none"
        self.role = "none"

        self.status = None
        self.show = None


class Room(object):
    """
    A Multi User Chat Room.

    An in memory object representing a MUC room from the perspective of
    a client.

    @ivar roomJID: The Room JID of the MUC room.
    @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

    @ivar nick: The nick name for the client in this room.
    @type nick: L{str}

    @ivar occupantJID: The JID of the occupant in the room. Generated from
        roomJID and nick.
    @type occupantJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

    @ivar locked: Flag signalling a locked room. A locked room first needs
        to be configured before it can be used. See
        L{MUCClientProtocol.getConfiguration} and
        L{MUCClientProtocol.configure}.
    @type locked: L{bool}
    """

    locked = False

    def __init__(self, roomJID, nick):
        """
        Initialize the room.
        """
        self.roomJID = roomJID
        self.setNick(nick)
        self.roster = {}

    def setNick(self, nick):
        self.occupantJID = jid.internJID("%s/%s" % (self.roomJID, nick))
        self.nick = nick

    def addUser(self, user):
        """
        Add a user to the room roster.

        @param user: The user object that is being added to the room.
        @type user: L{User}
        """
        self.roster[user.nick] = user

    def inRoster(self, user):
        """
        Check if a user is in the MUC room.

        @param user: The user object to check.
        @type user: L{User}
        """

        return user.nick in self.roster

    def getUser(self, nick):
        """
        Get a user from the room's roster.

        @param nick: The nick for the user in the MUC room.
        @type nick: L{str}
        """
        return self.roster.get(nick)

    def removeUser(self, user):
        """
        Remove a user from the MUC room's roster.

        @param user: The user object to check.
        @type user: L{User}
        """
        if self.inRoster(user):
            del self.roster[user.nick]


@implementer(IMUCClient)
class MUCClient(MUCClientProtocol):
    """
    Multi-User Chat client protocol.

    This is a subclass of L{XMPPHandler} and implements L{IMUCClient}.

    @ivar _rooms: Collection of occupied rooms, keyed by the bare JID of the
                  room. Note that a particular entity can only join a room once
                  at a time.
    @type _rooms: L{dict}
    """

    def __init__(self, reactor=None):
        MUCClientProtocol.__init__(self, reactor)

        self._rooms = {}

    def _addRoom(self, room):
        """
        Add a room to the room collection.

        Rooms are stored by the JID of the room itself. I.e. it uses the Room
        ID and service parts of the Room JID.

        @note: An entity can only join a particular room once.
        """
        roomJID = room.occupantJID.userhostJID()
        self._rooms[roomJID] = room

    def _getRoom(self, roomJID):
        """
        Grab a room from the room collection.

        This uses the Room ID and service parts of the given JID to look up
        the L{Room} instance associated with it.

        @type occupantJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """
        return self._rooms.get(roomJID)

    def _removeRoom(self, roomJID):
        """
        Delete a room from the room collection.
        """
        if roomJID in self._rooms:
            del self._rooms[roomJID]

    def _getRoomUser(self, stanza):
        """
        Lookup the room and user associated with the stanza's sender.
        """
        occupantJID = stanza.sender

        if not occupantJID:
            return None, None

        # when a user leaves a room we need to update it
        room = self._getRoom(occupantJID.userhostJID())
        if room is None:
            # not in the room yet
            return None, None

        # Check if user is in roster
        nick = occupantJID.resource
        user = room.getUser(nick)

        return room, user

    def unavailableReceived(self, presence):
        """
        Unavailable presence was received.

        If this was received from a MUC room occupant JID, that occupant has
        left the room.
        """

        room, user = self._getRoomUser(presence)

        if room is None or user is None:
            return

        room.removeUser(user)
        self.userLeftRoom(room, user)

    def availableReceived(self, presence):
        """
        Available presence was received.
        """

        room, user = self._getRoomUser(presence)

        if room is None:
            return

        if user is None:
            nick = presence.sender.resource
            user = User(nick, presence.entity)

        # Update user data
        user.role = presence.role
        user.affiliation = presence.affiliation
        user.status = presence.status
        user.show = presence.show

        if room.inRoster(user):
            self.userUpdatedStatus(room, user, presence.show, presence.status)
        else:
            room.addUser(user)
            self.userJoinedRoom(room, user)

    def groupChatReceived(self, message):
        """
        A group chat message has been received from a MUC room.

        There are a few event methods that may get called here.
        L{receivedGroupChat}, L{receivedSubject} or L{receivedHistory}.
        """
        room, user = self._getRoomUser(message)

        if room is None:
            return

        if message.subject:
            self.receivedSubject(room, user, message.subject)
        elif message.delay is None:
            self.receivedGroupChat(room, user, message)
        else:
            self.receivedHistory(room, user, message)

    def userJoinedRoom(self, room, user):
        """
        User has joined a MUC room.

        This method will need to be modified inorder for clients to
        do something when this event occurs.

        @param room: The room the user has joined.
        @type room: L{Room}

        @param user: The user that joined the MUC room.
        @type user: L{User}
        """
        pass

    def userLeftRoom(self, room, user):
        """
        User has left a room.

        This method will need to be modified inorder for clients to
        do something when this event occurs.

        @param room: The room the user has joined.
        @type room: L{Room}

        @param user: The user that left the MUC room.
        @type user: L{User}
        """
        pass

    def userUpdatedStatus(self, room, user, show, status):
        """
        User Presence has been received.

        This method will need to be modified inorder for clients to
        do something when this event occurs.
        """
        pass

    def receivedSubject(self, room, user, subject):
        """
        A (new) room subject has been received.

        This method will need to be modified inorder for clients to
        do something when this event occurs.
        """
        pass

    def receivedGroupChat(self, room, user, message):
        """
        A groupchat message was received.

        @param room: The room the message was received from.
        @type room: L{Room}

        @param user: The user that sent the message, or L{None} if it was a
            message from the room itself.
        @type user: L{User}

        @param message: The message.
        @type message: L{GroupChat}
        """
        pass

    def receivedHistory(self, room, user, message):
        """
        A groupchat message from the room's discussion history was received.

        This is identical to L{receivedGroupChat}, with the delayed delivery
        information (timestamp and original sender) in C{message.delay}. For
        anonymous rooms, C{message.delay.sender} is the room's address.

        @param room: The room the message was received from.
        @type room: L{Room}

        @param user: The user that sent the message, or L{None} if it was a
            message from the room itself.
        @type user: L{User}

        @param message: The message.
        @type message: L{GroupChat}
        """
        pass

    def join(self, roomJID, nick, historyOptions=None, password=None):
        """
        Join a MUC room by sending presence to it.

        @param roomJID: The JID of the room the entity is joining.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The nick name for the entitity joining the room.
        @type nick: L{str}

        @param historyOptions: Options for conversation history sent by the
            room upon joining.
        @type historyOptions: L{HistoryOptions}

        @param password: Optional password for the room.
        @type password: L{str}

        @return: A deferred that fires with the room when the entity is in the
            room, or with a failure if an error has occurred.
        """

        def cb(presence):
            """
            We have presence that says we joined a room.
            """
            if STATUS_CODE.ROOM_CREATED in presence.mucStatuses:
                room.locked = True

            return room

        def eb(failure):
            self._removeRoom(roomJID)
            return failure

        room = Room(roomJID, nick)
        self._addRoom(room)

        d = MUCClientProtocol.join(self, roomJID, nick, historyOptions, password)
        d.addCallbacks(cb, eb)
        return d

    def nick(self, roomJID, nick):
        """
        Change an entity's nick name in a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#changenick

        @param roomJID: The JID of the room, i.e. without a resource.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param nick: The new nick name within the room.
        @type nick: L{str}
        """

        def cb(presence):
            # Presence confirmation, change the nickname.
            room.setNick(nick)
            return room

        room = self._getRoom(roomJID)

        d = MUCClientProtocol.nick(self, roomJID, nick)
        d.addCallback(cb)
        return d

    def leave(self, roomJID):
        """
        Leave a MUC room.

        See: http://xmpp.org/extensions/xep-0045.html#exit

        @param roomJID: The Room JID of the room to leave.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}
        """

        def cb(presence):
            self._removeRoom(roomJID)

        d = MUCClientProtocol.leave(self, roomJID)
        d.addCallback(cb)
        return d

    def status(self, roomJID, show=None, status=None):
        """
        Change user status.

        See: http://xmpp.org/extensions/xep-0045.html#changepres

        @param roomJID: The Room JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param show: The availability of the entity. Common values are xa,
            available, etc
        @type show: L{str}

        @param status: The current status of the entity.
        @type status: L{str}
        """
        room = self._getRoom(roomJID)
        d = MUCClientProtocol.status(self, roomJID, show, status)
        d.addCallback(lambda _: room)
        return d

    def destroy(self, roomJID, reason=None, alternate=None, password=None):
        """
        Destroy a room.

        @param roomJID: The JID of the room.
        @type roomJID: L{JID<twisted.words.protocols.jabber.jid.JID>}

        @param reason: The reason for the destruction of the room.
        @type reason: L{str}

        @param alternate: The JID of the room suggested as an alternate venue.
        @type alternate: L{JID<twisted.words.protocols.jabber.jid.JID>}

        """

        def destroyed(iq):
            self._removeRoom(roomJID)

        d = MUCClientProtocol.destroy(self, roomJID, reason, alternate)
        d.addCallback(destroyed)
        return d



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.generic}.
"""

from __future__ import division, absolute_import

from zope.interface.verify import verifyObject

from twisted.trial import unittest
from twisted.words.xish import domish
from twisted.words.protocols.jabber.jid import JID

from wokkel import generic
from wokkel.iwokkel import IDisco
from wokkel.test.helpers import XmlStreamStub

NS_VERSION = "jabber:iq:version"


class VersionHandlerTest(unittest.TestCase):
    """
    Tests for L{wokkel.generic.VersionHandler}.
    """

    def setUp(self):
        self.protocol = generic.VersionHandler("Test", "0.1.0")

    def test_interface(self):
        """
        L{generic.VersionHandler} implements {IDisco}.
        """
        verifyObject(IDisco, self.protocol)

    def test_onVersion(self):
        """
        Test response to incoming version request.
        """
        self.stub = XmlStreamStub()
        self.protocol.xmlstream = self.stub.xmlstream
        self.protocol.send = self.stub.xmlstream.send
        self.protocol.connectionInitialized()

        iq = domish.Element((None, "iq"))
        iq["from"] = "user@example.org/Home"
        iq["to"] = "example.org"
        iq["type"] = "get"
        iq.addElement((NS_VERSION, "query"))
        self.stub.send(iq)

        response = self.stub.output[-1]
        self.assertEquals("user@example.org/Home", response["to"])
        self.assertEquals("example.org", response["from"])
        self.assertEquals("result", response["type"])
        self.assertEquals(NS_VERSION, response.query.uri)
        elements = list(
            domish.generateElementsQNamed(response.query.children, "name", NS_VERSION)
        )
        self.assertEquals(1, len(elements))
        self.assertEquals("Test", str(elements[0]))
        elements = list(
            domish.generateElementsQNamed(response.query.children, "version", NS_VERSION)
        )
        self.assertEquals(1, len(elements))
        self.assertEquals("0.1.0", str(elements[0]))


class XmlPipeTest(unittest.TestCase):
    """
    Tests for L{wokkel.generic.XmlPipe}.
    """

    def setUp(self):
        self.pipe = generic.XmlPipe()

    def test_sendFromSource(self):
        """
        Send an element from the source and observe it from the sink.
        """

        def cb(obj):
            called.append(obj)

        called = []
        self.pipe.sink.addObserver('/test[@xmlns="testns"]', cb)
        element = domish.Element(("testns", "test"))
        self.pipe.source.send(element)
        self.assertEquals([element], called)

    def test_sendFromSink(self):
        """
        Send an element from the sink and observe it from the source.
        """

        def cb(obj):
            called.append(obj)

        called = []
        self.pipe.source.addObserver('/test[@xmlns="testns"]', cb)
        element = domish.Element(("testns", "test"))
        self.pipe.sink.send(element)
        self.assertEquals([element], called)


class StanzaTest(unittest.TestCase):
    """
    Tests for L{generic.Stanza}.
    """

    def test_fromElement(self):
        xml = """
        <message type='chat' from='other@example.org' to='user@example.org'/>
        """

        stanza = generic.Stanza.fromElement(generic.parseXml(xml))
        self.assertEqual("chat", stanza.stanzaType)
        self.assertEqual(JID("other@example.org"), stanza.sender)
        self.assertEqual(JID("user@example.org"), stanza.recipient)

    def test_fromElementChildParser(self):
        """
        Child elements for which no parser is defined are ignored.
        """
        xml = """
        <message from='other@example.org' to='user@example.org'>
          <x xmlns='http://example.org/'/>
        </message>
        """

        class Message(generic.Stanza):
            childParsers = {("http://example.org/", "x"): "_childParser_x"}
            elements = []

            def _childParser_x(self, element):
                self.elements.append(element)

        message = Message.fromElement(generic.parseXml(xml))
        self.assertEqual(1, len(message.elements))

    def test_fromElementChildParserAll(self):
        """
        Child elements for which no parser is defined are ignored.
        """
        xml = """
        <message from='other@example.org' to='user@example.org'>
          <x xmlns='http://example.org/'/>
        </message>
        """

        class Message(generic.Stanza):
            childParsers = {None: "_childParser"}
            elements = []

            def _childParser(self, element):
                self.elements.append(element)

        message = Message.fromElement(generic.parseXml(xml))
        self.assertEqual(1, len(message.elements))

    def test_fromElementChildParserUnknown(self):
        """
        Child elements for which no parser is defined are ignored.
        """
        xml = """
        <message from='other@example.org' to='user@example.org'>
          <x xmlns='http://example.org/'/>
        </message>
        """
        generic.Stanza.fromElement(generic.parseXml(xml))


class RequestTest(unittest.TestCase):
    """
    Tests for L{generic.Request}.
    """

    def setUp(self):
        self.request = generic.Request()

    def test_requestParser(self):
        """
        The request's child element is passed to requestParser.
        """
        xml = """
        <iq type='get'>
          <query xmlns='jabber:iq:version'/>
        </iq>
        """

        class VersionRequest(generic.Request):
            elements = []

            def parseRequest(self, element):
                self.elements.append((element.uri, element.name))

        request = VersionRequest.fromElement(generic.parseXml(xml))
        self.assertEqual([(NS_VERSION, "query")], request.elements)

    def test_toElementStanzaKind(self):
        """
        A request is an iq stanza.
        """
        element = self.request.toElement()
        self.assertIdentical(None, element.uri)
        self.assertEquals("iq", element.name)

    def test_toElementStanzaType(self):
        """
        The request has type 'get'.
        """
        self.assertEquals("get", self.request.stanzaType)
        element = self.request.toElement()
        self.assertEquals("get", element.getAttribute("type"))

    def test_toElementStanzaTypeSet(self):
        """
        The request has type 'set'.
        """
        self.request.stanzaType = "set"
        element = self.request.toElement()
        self.assertEquals("set", element.getAttribute("type"))

    def test_toElementStanzaID(self):
        """
        A request, when rendered, has an identifier.
        """
        element = self.request.toElement()
        self.assertNotIdentical(None, self.request.stanzaID)
        self.assertEquals(self.request.stanzaID, element.getAttribute("id"))

    def test_toElementRecipient(self):
        """
        A request without recipient, has no 'to' attribute.
        """
        self.request = generic.Request(recipient=JID("other@example.org"))
        self.assertEquals(JID("other@example.org"), self.request.recipient)
        element = self.request.toElement()
        self.assertEquals("other@example.org", element.getAttribute("to"))

    def test_toElementRecipientNone(self):
        """
        A request without recipient, has no 'to' attribute.
        """
        element = self.request.toElement()
        self.assertFalse(element.hasAttribute("to"))

    def test_toElementSender(self):
        """
        A request with sender, has a 'from' attribute.
        """
        self.request = generic.Request(sender=JID("user@example.org"))
        self.assertEquals(JID("user@example.org"), self.request.sender)
        element = self.request.toElement()
        self.assertEquals("user@example.org", element.getAttribute("from"))

    def test_toElementSenderNone(self):
        """
        A request without sender, has no 'from' attribute.
        """
        element = self.request.toElement()
        self.assertFalse(element.hasAttribute("from"))

    def test_timeoutDefault(self):
        """
        The default is no timeout.
        """
        self.assertIdentical(None, self.request.timeout)

    def test_stanzaTypeInit(self):
        """
        If stanzaType is passed in __init__, it overrides the class variable.
        """

        class SetRequest(generic.Request):
            stanzaType = "set"

        request = SetRequest(stanzaType="get")
        self.assertEqual("get", request.stanzaType)

    def test_stanzaTypeClass(self):
        """
        If stanzaType is not passed in __init__, the class variable is used.
        """

        class SetRequest(generic.Request):
            stanzaType = "set"

        request = SetRequest()
        self.assertEqual("set", request.stanzaType)
